/** @file config.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ config ]
*/

#ifndef _CONFIG_H_
#define _CONFIG_H_

#include <tchar.h>
#include "vm/vm.h"

#define FILE_VERSION	0x30

void init_config();
void load_config();
void save_config();
void release_config();

typedef struct config_st {
	int version1;	// config file version
	int version2;

	// recent files
#ifdef USE_CART
	_TCHAR initial_cart_path[_MAX_PATH];
	_TCHAR recent_cart_path[8][_MAX_PATH];
#endif
#ifdef USE_FD1
	_TCHAR initial_disk_path[_MAX_PATH];
#if defined(USE_FD8) || defined(USE_FD7)
	_TCHAR recent_disk_path[8][8][_MAX_PATH];
#elif defined(USE_FD6) || defined(USE_FD5)
	_TCHAR recent_disk_path[6][8][_MAX_PATH];
#else
	_TCHAR recent_disk_path[4][8][_MAX_PATH];
#endif
#endif
#ifdef USE_QUICKDISK
	_TCHAR initial_quickdisk_path[_MAX_PATH];
	_TCHAR recent_quickdisk_path[8][_MAX_PATH];
#endif
#ifdef USE_DATAREC
	_TCHAR initial_datarec_path[_MAX_PATH];
	_TCHAR recent_datarec_path[8][_MAX_PATH];
#endif
#ifdef USE_MEDIA
	_TCHAR initial_media_path[_MAX_PATH];
	_TCHAR recent_media_path[8][_MAX_PATH];
#endif
#ifdef USE_BINARY_FILE1
	_TCHAR initial_binary_path[_MAX_PATH];
	_TCHAR recent_binary_path[2][8][_MAX_PATH];
#endif

	// screen
	int window_mode;
	bool stretch_screen;
	bool cutout_screen;

	// sound
	int sound_frequency;
	int sound_latency;

	// virtual machine
	int cpu_power;
#ifdef USE_FD1
	bool ignore_crc;
#endif
#ifdef USE_DIPSWITCH
	uint8 dipswitch;
#endif
#ifdef USE_BOOT_MODE
	int boot_mode;		// MZ-800, PASOPIA, PC-8801MA, PC-98DO
#endif
#ifdef USE_CPU_CLOCK_LOW
	bool cpu_clock_low;	// PC-8801MA, PC-9801E, PC-9801VM, PC-98DO
#endif
#if defined(_HC80) || defined(_PASOPIA) || defined(_PC8801MA)
	int device_type;
#endif
#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	int monitor_type;
#endif
#ifdef _BML3MK5
	bool sync_irq;
#endif
#ifdef USE_SCANLINE
	bool scan_line;
#ifdef _BML3MK5
	int  afterimage;
	int  keepimage;
	uint8 use_direct3d;
	uint8 disptmg_skew;
	uint8 curdisp_skew;
	// uint8
	int volume;
	int fdd_type;

	// CONFIG_VERSION >= 0x05
	// bit0: use fdd5  bit1: use fdd3  bit2: use psg  bit3: use ex pia
	// bit4: use ex acia  bit5: use 9psg  bit6: use kanji
	int io_port;

	uint8 show_dlg;

	_TCHAR initial_printer_path[_MAX_PATH];
	_TCHAR initial_state_path[_MAX_PATH];

	// CONFIG_VERSION >= 0x06
	int fps_no;

	int disp_device_no;
	int screen_width;
	int screen_height;

	_TCHAR initial_autokey_path[_MAX_PATH];
	// support above params on cfg file
	_TCHAR msgboard_fontname[_MAX_PATH];
	_TCHAR rom_path[_MAX_PATH];

	bool realmode_datarec;

	//
	uint8 screen_video_size;

	// CONFIG_VERSION >= 0x07
	int beep_volume;
	int psg6_volume;
	int psg9_volume;
	int relay_volume;
	int cmt_volume;
	bool mute;
	bool beep_mute;
	bool psg6_mute;
	bool psg9_mute;
	bool relay_mute;
	bool cmt_mute;

	// CONFIG_VERSION >= 0x08
	_TCHAR printer_server_host[MAX_PRINTER][_MAX_PATH];
	int    printer_server_port[MAX_PRINTER];
	bool   printer_direct[MAX_PRINTER];

	int    comm_dipswitch[MAX_COMM];
	_TCHAR comm_server_host[MAX_COMM][_MAX_PATH];
	int    comm_server_port[MAX_COMM];
	bool   comm_server[MAX_COMM];
	bool   comm_connect[MAX_COMM];
#endif
#endif
#ifdef USE_SOUND_DEVICE_TYPE
	int sound_device_type;
#endif
} config_t;

extern config_t config;

#endif

