/** @file configbox.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.03.24 -

	@brief [ config box ]
*/
#include "configbox.h"
#include "emu.h"

static _TCHAR *rate_list[] = {
	_T("11025"), _T("22050"), _T("44100"), _T("48000")
};
static _TCHAR *bits_list[] = {
	_T("8"), _T("16")
};
static _TCHAR *skew_list[] = {
	_T("0"), _T("1"), _T("2")
};

ConfigBox::ConfigBox(HINSTANCE hInst, EMU *new_emu, config_t *new_cfg)
{
	hInstance = hInst;
	emu       = new_emu;
	pConfig   = new_cfg;

	fdd_type  = new_cfg->fdd_type;

	wav_reverse = 0;
	wav_half = 1;
	wav_correct = 0;
	wav_correct_type = 0;

	wav_sample_rate = 3;
	wav_sample_bits = 0;

	io_port = new_cfg->io_port;
}

ConfigBox::~ConfigBox()
{
	pConfig->fdd_type = fdd_type;
	pConfig->io_port = io_port;
}

INT_PTR ConfigBox::Show(HWND hWnd)
{
	return DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_CONFIGURE), hWnd, ConfigBoxProc, (LPARAM)this);
}

INT_PTR CALLBACK ConfigBox::ConfigBoxProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	HWND hCtrl;
	ConfigBox *myObj;

	switch (message)
	{
	case WM_INITDIALOG:
	{
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		myObj = (ConfigBox *)lParam;

		return myObj->onInitDialog(hDlg, message, wParam, lParam);
	}
	case WM_COMMAND:
		myObj = (ConfigBox *)GetWindowLongPtr(hDlg, DWLP_USER);

		if (LOWORD(wParam) == IDOK) {
			myObj->onOK(hDlg, message, wParam, lParam);
		}
		else if (LOWORD(wParam) == IDC_RADIO_NOFDD) {
			CheckDlgButton(hDlg, IDC_RADIO_3FDD, 0);
			CheckDlgButton(hDlg, IDC_RADIO_5FDD, 0);
			// I/O port
			CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 0);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT1);
			EnableWindow(hCtrl, true);
			CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 0);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT2);
			EnableWindow(hCtrl, true);
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_RADIO_3FDD) {
			CheckDlgButton(hDlg, IDC_RADIO_NOFDD, 0);
			CheckDlgButton(hDlg, IDC_RADIO_5FDD, 0);
			// I/O port
			CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 0);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT1);
			EnableWindow(hCtrl, false);
			CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 1);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT2);
			EnableWindow(hCtrl, false);
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_RADIO_5FDD) {
			CheckDlgButton(hDlg, IDC_RADIO_NOFDD, 0);
			CheckDlgButton(hDlg, IDC_RADIO_3FDD, 0);
			// I/O port
			CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 1);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT1);
			EnableWindow(hCtrl, false);
			CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 0);
			hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT2);
			EnableWindow(hCtrl, false);
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_RADIO_COSW)
		{
			CheckDlgButton(hDlg, IDC_RADIO_SINW, 0);
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_RADIO_SINW)
		{
			CheckDlgButton(hDlg, IDC_RADIO_COSW, 0);
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_CHK_CORRECT)
		{
			if (IsDlgButtonChecked(hDlg, IDC_CHK_CORRECT)) {
				hCtrl = GetDlgItem(hDlg, IDC_RADIO_COSW);
				EnableWindow(hCtrl, true);
				hCtrl = GetDlgItem(hDlg, IDC_RADIO_SINW);
				EnableWindow(hCtrl, true);
			} else {
				hCtrl = GetDlgItem(hDlg, IDC_RADIO_COSW);
				EnableWindow(hCtrl, false);
				hCtrl = GetDlgItem(hDlg, IDC_RADIO_SINW);
				EnableWindow(hCtrl, false);
			}
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_CHK_IOPORT1)
		{
			if (IsDlgButtonChecked(hDlg, IDC_CHK_IOPORT1)) {
				CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 0);
			}
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_CHK_IOPORT2)
		{
			if (IsDlgButtonChecked(hDlg, IDC_CHK_IOPORT2)) {
				CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 0);
			}
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_CHK_IOPORT6)
		{
			if (IsDlgButtonChecked(hDlg, IDC_CHK_IOPORT6)) {
				CheckDlgButton(hDlg, IDC_CHK_IOPORT7, 0);
			}
			return (INT_PTR)FALSE;
		}
		else if (LOWORD(wParam) == IDC_CHK_IOPORT7)
		{
			if (IsDlgButtonChecked(hDlg, IDC_CHK_IOPORT7)) {
				CheckDlgButton(hDlg, IDC_CHK_IOPORT6, 0);
			}
			return (INT_PTR)FALSE;
		}

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}

INT_PTR ConfigBox::onInitDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hCtrl;
	RECT re;

	// FDD type select
	CheckDlgButton(hDlg, IDC_RADIO_NOFDD, fdd_type == 0);
	CheckDlgButton(hDlg, IDC_RADIO_3FDD, fdd_type == 1);
	CheckDlgButton(hDlg, IDC_RADIO_5FDD, fdd_type == 2);

	hCtrl = GetDlgItem(hDlg, IDC_STATIC_ALLOW);
	re.left = 10; re.top = (pConfig->fdd_type + 1) * 15;
	re.right = re.left + 10; re.bottom = re.top + 15;
	MapDialogRect(hDlg, &re);
	SetWindowPos(hCtrl, HWND_BOTTOM, re.left, re.top, 10, 15, SWP_NOSIZE | SWP_NOZORDER);

	// Load wave file
	CheckDlgButton(hDlg, IDC_CHK_REVERSE, wav_reverse);
	CheckDlgButton(hDlg, IDC_CHK_HALFWAVE, wav_half);

	CheckDlgButton(hDlg, IDC_RADIO_COSW, wav_correct_type == 0);
	CheckDlgButton(hDlg, IDC_RADIO_SINW, wav_correct_type == 1);
	if (wav_correct == 1) {
		CheckDlgButton(hDlg, IDC_CHK_CORRECT, 1);
		hCtrl = GetDlgItem(hDlg, IDC_RADIO_COSW);
		EnableWindow(hCtrl, true);
		hCtrl = GetDlgItem(hDlg, IDC_RADIO_SINW);
		EnableWindow(hCtrl, true);
	} else {
		CheckDlgButton(hDlg, IDC_CHK_CORRECT, 0);
		hCtrl = GetDlgItem(hDlg, IDC_RADIO_COSW);
		EnableWindow(hCtrl, false);
		hCtrl = GetDlgItem(hDlg, IDC_RADIO_SINW);
		EnableWindow(hCtrl, false);
	}

	// Save wave file
	SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)rate_list[0]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_ADDSTRING, 1, (LPARAM)(LPCSTR)rate_list[1]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_ADDSTRING, 2, (LPARAM)(LPCSTR)rate_list[2]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_ADDSTRING, 3, (LPARAM)(LPCSTR)rate_list[3]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_SETCURSEL, wav_sample_rate, 0);
	SendDlgItemMessage(hDlg, IDC_COMBO_SBITS, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)bits_list[0]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SBITS, CB_ADDSTRING, 1, (LPARAM)(LPCSTR)bits_list[1]);
	SendDlgItemMessage(hDlg, IDC_COMBO_SBITS, CB_SETCURSEL, wav_sample_bits, 0);

	// I/O port address
	for (int i=0; i<7; i++) {
		CheckDlgButton(hDlg, IDC_CHK_IOPORT1 + i, (io_port & (1 << i)) != 0);
		hCtrl = GetDlgItem(hDlg, IDC_STATIC_IOPORT1 + i);
		ShowWindow(hCtrl, (pConfig->io_port & (1 << i)) != 0 ? SW_SHOWNORMAL : SW_HIDE);
	}
	if (fdd_type == 1) {
		// I/O port
		CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 0);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT1);
		EnableWindow(hCtrl, false);
		CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 1);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT2);
		EnableWindow(hCtrl, false);
	}
	else if (fdd_type == 2) {
		// I/O port
		CheckDlgButton(hDlg, IDC_CHK_IOPORT1, 1);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT1);
		EnableWindow(hCtrl, false);
		CheckDlgButton(hDlg, IDC_CHK_IOPORT2, 0);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_IOPORT2);
		EnableWindow(hCtrl, false);
	}

	// crtc
	SendDlgItemMessage(hDlg, IDC_COMBO_DISPTMG, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)skew_list[0]);
	SendDlgItemMessage(hDlg, IDC_COMBO_DISPTMG, CB_ADDSTRING, 1, (LPARAM)(LPCSTR)skew_list[1]);
	SendDlgItemMessage(hDlg, IDC_COMBO_DISPTMG, CB_ADDSTRING, 2, (LPARAM)(LPCSTR)skew_list[2]);
	SendDlgItemMessage(hDlg, IDC_COMBO_DISPTMG, CB_SETCURSEL, config.disptmg_skew, 0);
	SendDlgItemMessage(hDlg, IDC_COMBO_CURDISP, CB_ADDSTRING, 0, (LPARAM)(LPCSTR)skew_list[0]);
	SendDlgItemMessage(hDlg, IDC_COMBO_CURDISP, CB_ADDSTRING, 1, (LPARAM)(LPCSTR)skew_list[1]);
	SendDlgItemMessage(hDlg, IDC_COMBO_CURDISP, CB_ADDSTRING, 2, (LPARAM)(LPCSTR)skew_list[2]);
	SendDlgItemMessage(hDlg, IDC_COMBO_CURDISP, CB_SETCURSEL, config.curdisp_skew, 0);

	return (INT_PTR)TRUE;
}

INT_PTR ConfigBox::onOK(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	fdd_type = (IsDlgButtonChecked(hDlg, IDC_RADIO_3FDD) ? 1 : 0)
			 + (IsDlgButtonChecked(hDlg, IDC_RADIO_5FDD) ? 2 : 0);

	wav_reverse = IsDlgButtonChecked(hDlg, IDC_CHK_REVERSE);
	wav_half = IsDlgButtonChecked(hDlg, IDC_CHK_HALFWAVE);
	wav_correct = IsDlgButtonChecked(hDlg, IDC_CHK_CORRECT);
	wav_correct_type = IsDlgButtonChecked(hDlg, IDC_RADIO_SINW);

	wav_sample_rate = (int)SendDlgItemMessage(hDlg, IDC_COMBO_SRATE, CB_GETCURSEL, 0, 0);
	wav_sample_bits = (int)SendDlgItemMessage(hDlg, IDC_COMBO_SBITS, CB_GETCURSEL, 0, 0);

	// I/O port address
	for (int i=0; i<7; i++) {
		if (IsDlgButtonChecked(hDlg, IDC_CHK_IOPORT1 + i))  {
			io_port |= (1 << i);
		} else {
			io_port &= ~(1 << i);
		}
	}
	// crtc
	config.disptmg_skew = (uint8)SendDlgItemMessage(hDlg, IDC_COMBO_DISPTMG, CB_GETCURSEL, 0, 0);
	config.curdisp_skew = (uint8)SendDlgItemMessage(hDlg, IDC_COMBO_CURDISP, CB_GETCURSEL, 0, 0);

	return (INT_PTR)TRUE;
}

void ConfigBox::Update()
{
	pConfig->fdd_type = fdd_type;

	switch(pConfig->fdd_type) {
		case 0:
			emu->out_log(LOG_INFO,_T("Non FDD Mode"));
			break;
		case 1:
			emu->out_log(LOG_INFO,_T("3inch compact FDD Mode"));
			break;
		case 2:
			emu->out_log(LOG_INFO,_T("5.25inch mini FDD Mode"));
			break;
		default:
			break;
	}

	pConfig->io_port = io_port;
}

int ConfigBox::GetParam(ConfigBoxParamId id)
{
	int data = 0;

	switch(id) {
		case FddType:
			data = fdd_type;
			break;
		case WavReverse:
			data = wav_reverse;
			break;
		case WavHalf:
			data = wav_half;
			break;
		case WavCorrect:
			data = wav_correct;
			break;
		case WavCorrectType:
			data = wav_correct_type;
			break;
		case WavSampleRate:
			data = wav_sample_rate;
			break;
		case WavSampleBits:
			data = wav_sample_bits;
			break;
		case IOPort:
			data = io_port;
			break;
	}

	return data;
}
