/** @file keybindbox.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.03.31 -

	@brief [ keybind box ]
*/

#include "keybindbox.h"
#include "emu.h"

KeybindBox::KeybindBox(HINSTANCE hInst)
{
		hInstance = hInst;
		memset(vmkey_map, 0, sizeof(vmkey_map));
}

void KeybindBox::SetVmKeyMap(int tab_num, UINT16 *vmKeyMap, int size)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	for(int i=0; i<KBCTRL_MAX_LINES && i<size; i++) {
		vmkey_map[tab_num][i] = vmKeyMap[i];
	}
	vmkey_map_size[tab_num] = (size > KBCTRL_MAX_LINES) ? KBCTRL_MAX_LINES : size;
}

void KeybindBox::SetVmKey(int tab_num, int idx, UINT16 code)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vmkey_map[tab_num][idx] = code;
}

void KeybindBox::SetVkKeyMap(int tab_num, UINT32 *vkKeyMap)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vkkey_map[tab_num] = vkKeyMap;
}

void KeybindBox::SetVkKeyDefMap(int tab_num, UINT32 *vkKeyDefMap, int rows, int cols)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vkkey_defmap[tab_num] = vkKeyDefMap;
	vkkey_defmap_rows[tab_num] = rows;
	vkkey_defmap_cols[tab_num] = cols;
}

void KeybindBox::SetVkKeyPresetMap(int tab_num, UINT32 *vkKeyMap, int idx)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	if (idx >= 0 && idx < MAX_PRESET_NUM) {
		vkkey_preset_map[tab_num][idx] = vkKeyMap;
	}
}

void KeybindBox::SetTitleLabel(_TCHAR *vm, _TCHAR *vk, _TCHAR *vj)
{
	_tcscpy(vm_label, vm);
	_tcscpy(vk_label, vk);
	_tcscpy(vj_label, vj);
}

INT_PTR KeybindBox::Show(HWND hWnd)
{
	return DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_KEYBIND), hWnd, Proc, (LPARAM)this);
}

// o[W{bNX̃bZ[W nhłB
INT_PTR CALLBACK KeybindBox::Proc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	KeybindBox *myObj;
//	_TCHAR str[1000];
//	_stprintf(str,_T("m:%x w:%d l:%d\n"),message,wParam,lParam);
//	OutputDebugString(str);

	switch (message)
	{
	case WM_INITDIALOG:
	{
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		myObj = (KeybindBox *)lParam;

		return myObj->onInitDialog(hDlg, message, wParam, lParam);
	}
	// zC[}EX
	case WM_MOUSEWHEEL:
		myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
		if (myObj->selected_tabctrl == 1) {
			SendDlgItemMessage(hDlg, IDC_CUSTOM2, message, wParam, lParam);
		} else {
			SendDlgItemMessage(hDlg, IDC_CUSTOM1, message, wParam, lParam);
		}
		return (INT_PTR)TRUE;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case IDOK:
		case IDCANCEL:
			if (LOWORD(wParam) == IDOK) {
				myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
				myObj->onClickOk(hDlg);
			}
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;

		case IDC_BTN_LOAD_DEFAULT:
			myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
			return myObj->onClickLoadDefault(hDlg);

		case IDC_BTN_LOAD_PRESET1:
		case IDC_BTN_LOAD_PRESET2:
		case IDC_BTN_LOAD_PRESET3:
		case IDC_BTN_LOAD_PRESET4:
			myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
			return myObj->onClickLoadPreset(hDlg, LOWORD(wParam) - IDC_BTN_LOAD_PRESET1);

		case IDC_BTN_SAVE_PRESET1:
		case IDC_BTN_SAVE_PRESET2:
		case IDC_BTN_SAVE_PRESET3:
		case IDC_BTN_SAVE_PRESET4:
			myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
			return myObj->onClickSavePreset(hDlg, LOWORD(wParam) - IDC_BTN_SAVE_PRESET1);

		}
		break;
	case WM_NOTIFY:
		// ^u؂ւ
		{
			LPNMHDR lpNmHdr = (NMHDR *)lParam;
			int i;
			if (lpNmHdr->idFrom = IDC_TAB1) {
				switch (lpNmHdr->code) {
				case TCN_SELCHANGE:
					myObj = (KeybindBox *)GetWindowLongPtr(hDlg, DWLP_USER);
					i = TabCtrl_GetCurSel(lpNmHdr->hwndFrom);
					myObj->select_tabctrl(hDlg, i);
					break;
				}
			}
		}
		return (INT_PTR)TRUE;
		break;

	}
	return (INT_PTR)FALSE;
}

INT_PTR KeybindBox::onInitDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hCtrl;
	TCITEM tcitm;
	UINT32 code;

	selected_tabctrl = 0;

	hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
	kbctl[0] = KeyBindControl_GetPtr(hCtrl);
	hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
	kbctl[1] = KeyBindControl_GetPtr(hCtrl);

	kbctl[0]->Init(1);
	kbctl[1]->Init(2);

	kbctl[0]->MapDefaultVmKeyCode((UINT16 *)vmkey_map[0]);
	kbctl[1]->MapDefaultVmKeyCode((UINT16 *)vmkey_map[1]);

	for(int tab_num=0; tab_num<MAX_TAB_NUM; tab_num++) {
		for(int row=0; row<vmkey_map_size[tab_num]; row++) {
			if (vmkey_map[tab_num][row] < vkkey_defmap_rows[tab_num]) {
				for(int col=0; col<vkkey_defmap_cols[tab_num]; col++) {
					code = *(vkkey_map[tab_num] + vmkey_map[tab_num][row] * vkkey_defmap_cols[tab_num] + col);
					if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(row, col, code);
					else kbctl[tab_num]->SetVkKeyCode(row, col, code);
				}
			}
		}
		if (tab_num == 1) {
			code = *(vkkey_map[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
			CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
		}
	}
	kbctl[0]->SetTitleLabel(vm_label, vk_label);
	kbctl[1]->SetTitleLabel(vm_label, vj_label);


	// ^uRg[
	hCtrl = GetDlgItem(hDlg, IDC_TAB1);

	tcitm.mask = TCIF_TEXT;

	tcitm.pszText = _T("keyboard");
	TabCtrl_InsertItem(hCtrl , 0 , &tcitm);
	tcitm.pszText = _T("joypad");
	TabCtrl_InsertItem(hCtrl , 1 , &tcitm);
	TabCtrl_SetCurSel(hCtrl, selected_tabctrl);

	select_tabctrl(hDlg, selected_tabctrl);

	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickOk(HWND hDlg)
{
	UINT32 code;

	for(int tab_num=0; tab_num<MAX_TAB_NUM; tab_num++) {
		for(int row=0; row<vmkey_map_size[tab_num]; row++) {
			if (vmkey_map[tab_num][row] < vkkey_defmap_rows[tab_num]) {
				for(int col=0; col<vkkey_defmap_cols[tab_num]; col++) {
					code = kbctl[tab_num]->GetVkKeyCode(row, col);
					*(vkkey_map[tab_num] + vmkey_map[tab_num][row] * vkkey_defmap_cols[tab_num] + col) = code;
				}
			}
		}
		if (tab_num == 1) {
			*(vkkey_map[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num])
				= (IsDlgButtonChecked(hDlg, IDC_CHK_COMBI1) ? 1 : 0);
		}
	}
	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickLoadDefault(HWND hDlg)
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = *(vkkey_defmap[tab_num] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j);
				if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(i, j, code);
				else kbctl[tab_num]->SetVkKeyCode(i, j, code);
			}

		}
	}
	if (tab_num == 1) {
		code = *(vkkey_defmap[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
		CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
	}
	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickLoadPreset(HWND hDlg, int idx)
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = *(vkkey_preset_map[tab_num][idx] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j);
				if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(i, j, code);
				else kbctl[tab_num]->SetVkKeyCode(i, j, code);
			}

		}
	}
	if (tab_num == 1) {
		code = *(vkkey_preset_map[tab_num][idx] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
		CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
	}

	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickSavePreset(HWND hDlg, int idx)
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = kbctl[tab_num]->GetVkKeyCode(i, j);
				*(vkkey_preset_map[tab_num][idx] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j) = code;
			}

		}
	}
	if (tab_num == 1) {
		*(vkkey_preset_map[tab_num][idx] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num])
			= (IsDlgButtonChecked(hDlg, IDC_CHK_COMBI1) ? 1 : 0);
	}

	return (INT_PTR)TRUE;
}

void KeybindBox::select_tabctrl(HWND hDlg, int tab_num)
{
	HWND hCtrl;

	selected_tabctrl = tab_num;
	if (selected_tabctrl == 1) {
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
		ShowWindow(hCtrl, SW_HIDE);
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
		ShowWindow(hCtrl, SW_SHOW);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_COMBI1);
		ShowWindow(hCtrl, SW_SHOW);
	} else {
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
		ShowWindow(hCtrl, SW_SHOW);
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
		ShowWindow(hCtrl, SW_HIDE);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_COMBI1);
		ShowWindow(hCtrl, SW_HIDE);
	}
}
