/// @file keybindctrl.h
///
/// @brief L[R[htRg[
///
/// @author Sasaji
/// @date   2012.3.31
///

#ifndef _KEYBINDCONTROL_H_
#define _KEYBINDCONTROL_H_

#include <windows.h>
#include "common.h"

#define KBCTRL_MAX_LINES 128
#define KBCTRL_MAX_COLS  2

class KeyBindControl;

ATOM KeyBindControl_RegisterClass(HINSTANCE);
void KeyBindControl_UnregisterClass();

KeyBindControl *KeyBindControl_GetPtr(HWND);

typedef struct codecols_st {
	int     col;
	HWND    hEdt;
	WNDPROC proc;
	UINT32  vk_keycode;
	UINT32  vk_prev_keycode;
} codecols_t;

typedef struct codetable_st {
	int        row;
	HWND       hSta1;
	codecols_t cols[KBCTRL_MAX_COLS];
	UINT16     vm_keycode;
	bool       enabled;
} codetable_t;

/// @brief L[R[htRg[
class KeyBindControl
{
private:
	HINSTANCE hInstance;
	HWND hMainCtrl;
	HWND hStaticMain;

	UINT8	usekey_type;

	HWND       hTitles[KBCTRL_MAX_COLS + 1];
	codetable_t table[KBCTRL_MAX_LINES];

//	UINT8 joy_status[2];	// joystick #1, #2 (b0 = up, b1 = down, b2 = left, b3 = right, b4-b7 = trigger #1-#4
	int joyid_map[2];
	UINT32 joy_xmin[2], joy_xmax[2];
	UINT32 joy_ymin[2], joy_ymax[2];

	int cell_height;
	int cell_width;
	int padding;
	int margin;

	SCROLLINFO vScrMain;

	void create_children(HWND hWnd);
	void setfont_children(HWND hWnd, WPARAM wParam);
	void destroy_children(HWND hWnd);
	void dispose_children(HWND hWnd);
	LRESULT ctlcolor_children(HWND hWnd, WPARAM wParam, LPARAM lParam);
	void move_children(HWND hWnd, int x, int y);

	void init_joypad(HWND hWnd);
	void term_joypad(HWND hWnd);

	bool get_vmkeylabel(int code, _TCHAR *label);
	bool get_vkkeylabel(UINT32 code, _TCHAR *label);
	bool get_vkjoylabel(UINT32 code, _TCHAR *label);
	UINT32 translate_vkkey(UINT32 code);

	void set_scroll_bar_range(HWND hWnd, int y);
	void set_scroll_bar_pos(HWND hWnd, int dy);

	friend LRESULT CALLBACK CustomCtrlMainProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	friend LRESULT CALLBACK CustomCtrlEditProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
public:
	KeyBindControl(HINSTANCE hInst, HWND hWnd);
	~KeyBindControl();
	void Init(UINT8 flags);

	LRESULT Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT MainProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT EditProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

	void SetTitleLabel(const _TCHAR *vmlabel, const _TCHAR *vklabel);

	bool   SetVmKeyCode(int idx, UINT16 code);
	UINT16 GetVmKeyCode(int idx);
	bool   SetVkKeyCode(int idx, int col, UINT32 code);
	bool   SetVkKeyCode(codecols_t *obj, UINT32 code);
	bool   SetVkJoyCode(int idx, int col, UINT32 code);
	bool   SetVkJoyCode(codecols_t *obj, UINT32 code);
	bool   UndoVkKeyCode(codecols_t *obj);
	UINT32 GetVkKeyCode(int idx, int col);
	bool   MapDefaultVmKeyCode(UINT16 *tbl);
};


#endif
