/** @file ledbox.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.08.08 -

	@brief [ led box ]
*/
#include "ledbox.h"
#include "emu.h"

LedBox::LedBox(HINSTANCE hInst) : DibSection()
{
	hInstance = hInst;
	hLedBox = NULL;

	for(int i=0; i<LED_TYPE_END; i++) {
		ZeroMemory(&led[i], sizeof(led_t));
		led[i].hBmp = NULL;
		led[i].hDC = NULL;
	}
	ZeroMemory(led_pt, sizeof(led_pt));

	flag_old = -1;
	dist.x = 0; dist.y = 0;
	mode = 0;
	dist_def[0].x = 2; dist_def[0].y = 2;
	dist_def[1].x = 2; dist_def[1].y = 2;
	dist_set[0].x = 2; dist_set[0].y = 2;
	dist_set[1].x = 2; dist_set[1].y = 2;

	visible = false;
	inside = false;
}

LedBox::~LedBox()
{
	for(int i=0; i<LED_TYPE_END; i++) {
		if (led[i].hDC) DeleteDC(led[i].hDC);
		if (led[i].hBmp) DeleteObject(led[i].hBmp);
	}
}

// ʏ
void LedBox::InitScreen(HDC hdc, int width, int height)
{
	// LEDwi
	if (!create_bitmap(hdc, &led[LED_TYPE_BASE], IDB_KBLEDBOX)) return;

	win_sz.cx = led[LED_TYPE_BASE].sz.cx;
	win_sz.cy = led[LED_TYPE_BASE].sz.cy;

	// Dib쐬
	if (!CreateDib(hdc, win_sz.cx, win_sz.cy)) return;

	// LEDp[c쐬
	enable = false;
	// horiz
	if (!create_bitmap(hdc, &led[LED_TYPE_HR], IDB_LED_R_H_RED)) return;
	if (!create_bitmap(hdc, &led[LED_TYPE_HG], IDB_LED_R_H_GREEN)) return;
	// vert
	if (!create_bitmap(hdc, &led[LED_TYPE_VR], IDB_LED_RV_H_RED)) return;
	if (!create_bitmap(hdc, &led[LED_TYPE_VG], IDB_LED_RV_H_GREEN)) return;
	// datarec
	if (!create_bitmap(hdc, &led[LED_TYPE_RR], IDB_LED_RR_H_GREEN)) return;	// play led
	if (!create_bitmap(hdc, &led[LED_TYPE_CL], IDB_LED_C_H_RED)) return;	// rec led

	// ʒuݒ
	led_pt[LED_POS_BASE].x = 0; led_pt[LED_POS_BASE].y = 0;
	// keyboard led
	led_pt[LED_POS_KATAKANA].x = 3; led_pt[LED_POS_KATAKANA].y = 3;
	led_pt[LED_POS_HIRAGANA].x = 3; led_pt[LED_POS_HIRAGANA].y = 11;
	led_pt[LED_POS_CAPS].x = 36; led_pt[LED_POS_CAPS].y = 3;
	// datarec led
	led_pt[LED_POS_PLAY].x = 65; led_pt[LED_POS_PLAY].y = 4;
	led_pt[LED_POS_REC].x = 98; led_pt[LED_POS_REC].y = 4;
	// fdd 1
	led_pt[LED_POS_FDD1].x = 130; led_pt[LED_POS_FDD1].y = 3;
	// fdd 2
	led_pt[LED_POS_FDD2].x = 155; led_pt[LED_POS_FDD2].y = 3;

	// 
	enable = true;
	copy_led(led_pt[LED_POS_BASE], led[LED_TYPE_BASE]);
}

bool LedBox::create_bitmap(HDC hdc, led_t *l, DWORD id)
{
	BITMAPINFO bInfo;
	ZeroMemory(&bInfo, sizeof(bInfo));
	bInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

	l->hBmp = LoadBitmap(hInstance,MAKEINTRESOURCE(id));
	if (l->hBmp == NULL) return false;
	l->hDC = CreateCompatibleDC(hdc);
	if (l->hDC == NULL) return false;
	SelectObject(l->hDC, l->hBmp);
	GetDIBits(hdc, l->hBmp, 0, 1, NULL, &bInfo, DIB_RGB_COLORS);
	l->sz.cx = bInfo.bmiHeader.biWidth;
	l->sz.cy = bInfo.bmiHeader.biHeight;

	return true;
}

void LedBox::Create(HWND hWnd)
{
	if (!enable) return;

	hLedBox = CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_KBLEDBOX), hWnd, LedBoxProc, (LPARAM)this);
}

void LedBox::Show(int flag)
{
	visible = ((flag & 1) != 0);
	inside = ((flag & 8) != 0);

	if (hLedBox) {
		ShowWindow(hLedBox, (visible && !inside) ? SW_SHOWNORMAL : SW_HIDE);
		HWND hWnd = GetParent(hLedBox);
		SetActiveWindow(hWnd);
	}
}

void LedBox::Move()
{
	WINDOWINFO wi;
	WINDOWINFO wid;

	if (hLedBox) {
		HWND hWnd = GetParent(hLedBox);
		GetWindowInfo(hWnd, &wi);
		GetWindowInfo(hLedBox, &wid);
		int w = wid.rcWindow.right - wid.rcWindow.left;
		int h = wid.rcWindow.bottom - wid.rcWindow.top;
		int x = wi.rcClient.left + dist.x;
		int y = wi.rcClient.top + dist.y;
		MoveWindow(hLedBox, x, y, w, h, true);
	}
}

void LedBox::MoveInPlace()
{
	WINDOWINFO wi;
	WINDOWINFO wid;

	if (hLedBox) {
		HWND hWnd = GetParent(hLedBox);
		GetWindowInfo(hWnd, &wi);
		GetWindowInfo(hLedBox, &wid);
		int w = wid.rcWindow.right - wid.rcWindow.left;
		int h = wid.rcWindow.bottom - wid.rcWindow.top;
		dist.x = dist_set[mode].x;
		dist.y = dist_set[mode].y;
		int x = wi.rcClient.left + dist.x;
		int y = wi.rcClient.top + dist.y;
		MoveWindow(hLedBox, x, y, w, h, true);
	}
}

INT_PTR CALLBACK LedBox::LedBoxProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	HWND hDCtl;
	WINDOWINFO wi;
	WINDOWINFO wid;
	static POINT pStart;
	LedBox *myObj;

	switch (message)
	{
	case WM_INITDIALOG:
	{
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);

		return (INT_PTR)TRUE;
	}
	case WM_LBUTTONDOWN:
		pStart.x = GET_X_LPARAM(lParam);
		pStart.y = GET_Y_LPARAM(lParam);
		return (INT_PTR)FALSE;
	case WM_MOUSEMOVE:
		myObj = (LedBox *)GetWindowLongPtr(hDlg, DWLP_USER);

		if (wParam & MK_LBUTTON)
		{
			hDCtl = GetParent(hDlg);
			GetWindowInfo(hDCtl, &wi);
			GetWindowInfo(hDlg, &wid);
			int w = wid.rcWindow.right - wid.rcWindow.left;
			int h = wid.rcWindow.bottom - wid.rcWindow.top;
			int x = GET_X_LPARAM(lParam) - pStart.x + wid.rcWindow.left;
			int y = GET_Y_LPARAM(lParam) - pStart.y + wid.rcWindow.top;
			myObj->dist.x = wid.rcWindow.left - wi.rcClient.left;
			myObj->dist.y = wid.rcWindow.top - wi.rcClient.top;
			MoveWindow(hDlg, x, y, w, h, true);
		}
		return (INT_PTR)FALSE;
	case WM_LBUTTONUP:
		hDCtl = GetParent(hDlg);
		SetActiveWindow(hDCtl);
		return (INT_PTR)FALSE;
	case WM_COMMAND:
		return (INT_PTR)TRUE;
		break;
	case WM_PAINT:
		{
			myObj = (LedBox *)GetWindowLongPtr(hDlg, DWLP_USER);
			PAINTSTRUCT ps;
			HDC hdc;
			hdc = BeginPaint(hDlg, &ps);
			BitBlt(hdc, 0, 0, myObj->win_sz.cx, myObj->win_sz.cy, myObj->hMainDib, 0, 0, SRCCOPY);
			EndPaint(hDlg, &ps);
			return (INT_PTR)FALSE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}

void LedBox::Update(int flag)
{
	if (!enable) return;

	if (flag != flag_old) {
		copy_led(led_pt[LED_POS_BASE], led[LED_TYPE_BASE]);

		if (flag & 0x0001) {
			copy_led(led_pt[LED_POS_KATAKANA], led[LED_TYPE_HR]);
		}
		if (flag & 0x0002) {
			copy_led(led_pt[LED_POS_HIRAGANA], led[LED_TYPE_HG]);
		}
		if (!(flag & 0x0004)) {
			copy_led(led_pt[LED_POS_CAPS], led[LED_TYPE_VR]);
		}

		if (flag & 0x0010) {
			copy_led(led_pt[LED_POS_PLAY], led[LED_TYPE_RR]);
		}
		if (flag & 0x0020) {
			copy_led(led_pt[LED_POS_REC], led[LED_TYPE_CL]);
		}

		if ((flag & 0x1100) == 0x0100) {
			copy_led(led_pt[LED_POS_FDD1], led[LED_TYPE_VG]);
		} else if ((flag & 0x1100) == 0x1100) {
			copy_led(led_pt[LED_POS_FDD1], led[LED_TYPE_VR]);
		}

		if ((flag & 0x2200) == 0x0200) {
			copy_led(led_pt[LED_POS_FDD2], led[LED_TYPE_VG]);
		} else if ((flag & 0x2200) == 0x2200) {
			copy_led(led_pt[LED_POS_FDD2], led[LED_TYPE_VR]);
		}

		if (hLedBox && visible && !inside) {
			InvalidateRect(hLedBox, NULL, TRUE);
			UpdateWindow(hLedBox);
		}
	}

	flag_old = flag;
}

void LedBox::copy_led(POINT &pt, led_t &l)
{
	BitBlt(hMainDib, pt.x, pt.y, l.sz.cx, l.sz.cy, l.hDC, 0, 0, SRCCOPY); 
}

void LedBox::SetMode(int val)
{
	val = (val < 0) ? 0 : (val > 1 ? 1 : val);
	dist_set[mode].x = dist.x;
	dist_set[mode].y = dist.y;
	mode = val;
	dist.x = dist_set[mode].x;
	dist.y = dist_set[mode].y;
}
#if 0
void LedBox::SetQuadrant()
{
	if (hLedBox) {
		HWND hWnd = GetParent(hLedBox);
		GetWindowInfo(hWnd, &wi);
		GetWindowInfo(hLedBox, &wid);
	}
}
#endif

void LedBox::SetPos(int cx, int cy)
{
	parent_pt.x = cx;
	parent_pt.y = cy;
}

void LedBox::Draw(HDC hdc)
{
	if (!visible || !inside || !enable) return;

	// copy to main context
	BitBlt(hdc, parent_pt.x, parent_pt.y, win_sz.cx, win_sz.cy, hMainDib, 0, 0, SRCCOPY);
}

void LedBox::Draw(LPDIRECT3DSURFACE9 suf)
{
	if (!visible || !inside || !enable || !suf) return;

	HDC hdc;
	HRESULT hre = suf->GetDC(&hdc);
	if (hre == D3D_OK) {
		// copy to main context
		BitBlt(hdc, parent_pt.x, parent_pt.y, win_sz.cx, win_sz.cy, hMainDib, 0, 0, SRCCOPY);
	}
	suf->ReleaseDC(hdc);
}
