/** @file msgboard.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.26 -

	@brief [ message board ]
*/
#include "msgboard.h"
#include "common.h"
#include "emu.h"
#include "config.h"

MsgBoard::MsgBoard(HWND hWnd, EMU *pEmu) : DibSection()
{
	visible = true;

	info.hFont = NULL;
	msg.hFont  = NULL;

	hPen   = NULL;
	hBrush = NULL;

	fg = 0;
	bg = 0;

	szWin.cx = 1;
	szWin.cy = 1;

	emu = pEmu;
}

MsgBoard::~MsgBoard()
{
	if (enable) {
		if (info.hFont) DeleteObject(info.hFont);
		if (msg.hFont)  DeleteObject(msg.hFont);
		if (hBrush)     DeleteObject(hBrush);
		if (hPen)       DeleteObject(hPen);
	}
}

// ʏ
void MsgBoard::InitScreen(HDC hdc, int width, int height)
{
	LOGFONT lf;

	// F
	fg = RGB(0x00, 0xc0, 0x80);
	bg = RGB(0x00, 0x40, 0x00);

	// EBhETCY
	szWin.cx = width;
	szWin.cy = height;

	// Dib쐬
	CreateDib(hdc, width, 64);

	if (enable) {
		// `py̐ݒ
		hPen = CreatePen(PS_SOLID, 1, fg);
		SelectObject(hMainDib, hPen);

		// `puV̐ݒ
		hBrush = CreateSolidBrush(bg);
		SelectObject(hMainDib, hBrush);

		// bZ[WptHg̐ݒ
		set_sys_font(&lf, 0, 14, FW_DONTCARE);
		msg.hFont = CreateFontIndirect(&lf);
		// gݒ
		SetRect(&msg.re, 0, 0, width, 31);
		// \ʒu
		msg.pt.x = 0; msg.pt.y = 0; info.place = 2;

		// ptHg̐ݒ
		set_sys_font(&lf, 0, 24, FW_BOLD);
		info.hFont = CreateFontIndirect(&lf);
		// gݒ
		SetRect(&info.re, 0, 32, width, 63);
		// \ʒu
		info.pt.x = 0; info.pt.y = 0; info.place = 1;
	}
	if (msg.hFont== NULL || info.hFont == NULL) {
		enable = false;
	}
	if (enable) {
		emu->out_log(LOG_INFO , _T("MsgBoard: OK"));
	} else {
		emu->out_log(LOG_ERROR, _T("MsgBoard: Failed"));
	}
}

// o
void MsgBoard::draw(HDC hdc, msg_data_t &data)
{
	POINT pt;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// ʒǔvZ
		pt = data.pt;
		if (data.place & 1) {
			pt.x += szWin.cx - data.sz.cx;
		}
		if (data.place & 2) {
			pt.y += szWin.cy - data.sz.cy;
		}

		// CReLXgɃbZ[WRs[
		BitBlt(hdc, pt.x, pt.y,
			data.sz.cx, data.sz.cy,
			hMainDib, data.re.left, data.re.top, SRCCOPY);
	}
}
void MsgBoard::draw(LPDIRECT3DSURFACE9 suf, msg_data_t &data)
{
	RECT  reSrc;
	RECT  reDst;

	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// ʒǔvZ
		if (data.place & 1) {
			reDst.left = szWin.cx + data.pt.x - data.sz.cx;
			reDst.right = szWin.cx + data.pt.x;
		} else {
			reDst.left = data.pt.x;
			reDst.right = data.pt.x + data.sz.cx;
		}
		if (data.place & 2) {
			reDst.top = szWin.cy + data.pt.y - data.sz.cy;
			reDst.bottom = szWin.cy + data.pt.y;
		} else {
			reDst.top = data.pt.y;
			reDst.bottom = data.pt.y + data.sz.cy;
		}
		SetRect(&reSrc, data.re.left, data.re.top, data.re.left + data.sz.cx, data.re.top + data.sz.cy);

		// CReLXgɃbZ[WRs[
#ifdef USE_DIRECT3DX
		D3DXLoadSurfaceFromMemory(suf,
			NULL,
			&reDst,
			pMainBuf,
			D3DFMT_X8R8G8B8,
			szWin.cx * 4,
			NULL,
			&reSrc,
			D3DX_FILTER_NONE,
			0);
#else
		if (suf) {
			HDC hdc;
			HRESULT hre = suf->GetDC(&hdc);
			if (hre == D3D_OK) {
				BitBlt(hdc, reDst.left, reDst.top,
					data.sz.cx, data.sz.cy,
					hMainDib, data.re.left, data.re.top, SRCCOPY);
			}
			suf->ReleaseDC(hdc);
		}
#endif
	}
}

void MsgBoard::Draw(HDC hdc)
{
	if (!enable || !visible) return;

	draw(hdc, msg);
	draw(hdc, info);
}

void MsgBoard::Draw(LPDIRECT3DSURFACE9 suf)
{
	if (!enable || !visible) return;

	draw(suf, msg);
	draw(suf, info);
}

// obNobt@ɕ`
void MsgBoard::draw_text(msg_data_t &data)
{
	SIZE sz;
	int len = 0;

	if (enable && !data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// wiF
		SetBkColor(hMainDib, bg);
		// F
		SetTextColor(hMainDib, fg);

		// g`
		FillRect(hMainDib, &data.re, hBrush);

		// tHgݒ
		SelectObject(hMainDib, data.hFont);

		// `
		len = (int)_tcslen(it->msg);
		TextOut(hMainDib, data.re.left + 2, data.re.top + 2, it->msg, len);

		// ̘̕gvZ
		GetTextExtentPoint32(hMainDib, it->msg, len, &sz);
		data.sz.cx = sz.cx + 4;
		data.sz.cy = sz.cy + 4;
	}
}

// bZ[Wݒ
void MsgBoard::Set(msg_data_t &data, const _TCHAR *str, int sec)
{
	item_t itm;

	_tcsncpy(itm.msg, str, 1023);
	itm.cnt = (60 * sec);

	data.lists.push_front(itm);
	if (data.lists.size() > 10) {
		data.lists.pop_back();
	}

	if (enable) {
		draw_text(data);
	}
}
void MsgBoard::SetMessage(const _TCHAR *str, int sec)
{
	Set(msg, str, sec);
}
void MsgBoard::SetInfo(const _TCHAR *str, int sec)
{
	Set(info, str, sec);
}
void MsgBoard::SetMessageF(const _TCHAR *format, ...)
{
	_TCHAR buf[1024];

	va_list ap;

	va_start(ap, format);
	_vstprintf(buf, format, ap);
	va_end(ap);

	SetMessage(buf);
}
void MsgBoard::SetInfoF(const _TCHAR *format, ...)
{
	_TCHAR buf[1024];

	va_list ap;

	va_start(ap, format);
	_vstprintf(buf, format, ap);
	va_end(ap);

	SetInfo(buf);
}

// EBhETCYݒ
void MsgBoard::SetSize(int width, int height)
{
	szWin.cx = width;
	szWin.cy = height;
}

// bZ[W`ʒuݒ
void MsgBoard::SetMessagePos(int cx, int cy, int place)
{
	msg.pt.x = cx;
	msg.pt.y = cy;
	msg.place = place;
}
void MsgBoard::SetInfoPos(int cx, int cy, int place)
{
	info.pt.x = cx;
	info.pt.y = cy;
	info.place = place;
}

// bZ[W폜
void MsgBoard::Delete(msg_data_t &data, const _TCHAR *str)
{
	bool redraw = false;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();
		list_t::iterator it_next;
		while(it != data.lists.end())	{
			it_next = it;
			it_next++;

			if (_tcscmp(str, it->msg) == 0) {
				data.lists.erase(it);
				redraw = true;
			}
			it = it_next;
		}
	}
	if (redraw) {
		draw_text(data);
	}
}
void MsgBoard::DeleteMessage(const _TCHAR *str)
{
	Delete(msg, str);
}
void MsgBoard::DeleteInfo(const _TCHAR *str)
{
	Delete(info, str);
}

// JEg_E
void MsgBoard::count_down(msg_data_t &data)
{
	bool redraw = false;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();
		list_t::iterator it_next;
		while(it != data.lists.end())	{
			it_next = it;
			it_next++;

			if (it->cnt > 0) it->cnt--;
			if (it->cnt == 0) {
				data.lists.erase(it);
				redraw = true;
			}
			it = it_next;
		}
	}
	if (redraw) {
		draw_text(data);
	}
}
void MsgBoard::CountDown(void)
{
	count_down(msg);
	count_down(info);
}

// ʕ\pVXetHg̐ݒ
void MsgBoard::set_sys_font(LPLOGFONT plf, int w, int h, long weight)
{

	memset(plf, 0, sizeof(LOGFONT));
    plf->lfHeight = h;				// Z܂͕̍
    plf->lfWidth = w;					// ϕ
//    plf->lfEscapement = 900;			// ̕XƂ̊px 1/10PʂŎwB
    plf->lfEscapement = 0;			// ̕XƂ̊px 1/10PʂŎwB
    plf->lfOrientation = plf->lfEscapement;			// x[XCXƂ̊px
    plf->lfWeight = weight;		// tHg̑
    plf->lfItalic = 0;				// C^bN̎w
    plf->lfUnderline = 0;			// tw
    plf->lfStrikeOut = 0;				// łtw
    plf->lfCharSet = DEFAULT_CHARSET;	// LN^Zbg
    plf->lfOutPrecision = OUT_DEFAULT_PRECIS;		// o͐x
    plf->lfClipPrecision = CLIP_DEFAULT_PRECIS;	// NbsO̐x
    plf->lfQuality = DEFAULT_QUALITY;	// o͕i
    plf->lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;	// sb`ƃt@~
	// tHg
	_tcsncpy(plf->lfFaceName, config.msgboard_fontname, LF_FACESIZE);
}
