/** @file bml3mk5.h

	HITACHI BASIC MASTER LEVEL3 MARK5 Emulator
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ virtual machine ]
*/

#ifndef _BML3MK5_H_
#define _BML3MK5_H_

#define FRAME_SPLIT_NUM	2

#define DEVICE_NAME		"HITACHI BASIC MASTER LEVEL3 MARK5"
#define CONFIG_NAME		"bml3mk5"
#define CLASS_NAME      "BML3MK5"
#define CONFIG_VERSION		0x08

// device informations for virtual machine
#define FRAMES_PER_10SECS	600
#define FRAMES_PER_SEC		60
#define LINES_PER_FRAME 	262
#define CHARS_PER_LINE		64

#define CPU_CLOCKS		1008000

//#define SCREEN_WIDTH		640
#define SCREEN_WIDTH		768
//#define SCREEN_HEIGHT		480
#define SCREEN_HEIGHT		512
#define MIN_WINDOW_WIDTH		640
#define MIN_WINDOW_HEIGHT		480
#define MAX_WINDOW_WIDTH		768
#define MAX_WINDOW_HEIGHT		512

// device informations for win32
#define USE_SPECIAL_RESET
#define USE_DATAREC
// #define USE_ALT_F10_KEY
#define USE_AUTO_KEY		(5 + FRAME_SPLIT_NUM)
#define USE_AUTO_KEY_RELEASE	(USE_AUTO_KEY + 3)
#define USE_AUTO_KEY_CAPS
#define USE_SCANLINE
#define USE_DIPSWITCH
#define DIPSWITCH_DEFAULT 0x03
#define HAS_AY_3_8912

#define USE_PRINTER
#define MAX_PRINTER		3
#define USE_LIGHTPEN
#define USE_JOYSTICK

#define USE_FD1
#define USE_FD2
#define HAS_MB8876
#define MAX_DRIVE		4

#define USE_SOCKET
#define MAX_COMM		2

#define RESUME_FILE_HEADER "RESUME_BML3MK5"
#define RESUME_FILE_VERSION 1
#define RESUME_FILE_REVISION 1

#define KEYBIND_KEYS	130
#define KEYBIND_ASSIGN	2
#define KEYBIND_PRESETS	4

//#define _FDC_DEBUG_LOG

#include "../../common.h"

class EMU;
class DEVICE;
class EVENT;

class HD46505;
class YM2203;
class MC6809;

class BOARD;
class PIA;
class ACIA;
class DISPLAY;
class KEYBOARD;
class MEMORY;
class SOUND;
class COMM;
class CMT;
class PRINTER;
class TIMER;
class KANJI;
class VIA;
class PSGC;
class PSG9C;
#ifdef USE_FD1
class MC6843;
class MB8866;
class FLOPPY;
#endif

/// for sleep/resume the machine
#pragma pack(1)
typedef struct vm_state_header_st {
	char header[16];
	uint16 version;
	uint16 revision;
	uint32 param;
	char reserved[8];
} vm_state_header_t;
#pragma pack()

/**
	@brief virtual machine for BML3MK5
*/
class VM
{
protected:
	EMU* emu;

	/// @name devices
	//@{
	EVENT* event;

	HD46505* crtc;
	YM2203* psg[2];
	YM2203* psg9[3];
	MC6809* cpu;

	BOARD* board;
	PIA* pia;
	ACIA* acia;
	DISPLAY* display;
	KEYBOARD* key;
	MEMORY* memory;
	SOUND* sound;
	COMM* comm[MAX_COMM];
	CMT* cmt;
	PRINTER* printer[MAX_PRINTER];
	TIMER* timer;
	KANJI* kanji;
	PIA* psg_pia;
	PSGC* psgc;
	VIA* psg9_via;
	PSG9C* psg9c;
#ifdef USE_FD1
	MC6843* fdc3;
	MB8866* fdc5;
	FLOPPY* fdd;
#endif
	PIA* pia_ex;
	ACIA* acia_ex;

	DEVICE* dummy;
	//@}

public:
	DEVICE* first_device;
	DEVICE* last_device;

public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	VM(EMU* parent_emu);
	~VM();
	//@}
	// ----------------------------------------
	// for emulation class
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	void run(int split_num);
	double frame_rate();
	bool now_skip();
	//@}
	/// @name draw screen
	//@{
	void draw_screen();
	int  update_led();
	//@}
	/// @name sound generation
	//@{
	void initialize_sound(int rate, int samples);
	uint16* create_sound(int* extra_frames);

	//@}
	/// @name input event from emu
	//@{
	void key_down(int code, bool repeat);
	void key_up(int code);
	void init_keybind_dialog();
	//@}
	/// @name socket event from emu
	//@{
	void network_connected(int ch);
	void network_disconnected(int ch);
	void network_writeable(int ch);
	void network_readable(int ch);
	void network_accepted(int ch, int new_ch);
	uint8* get_sendbuffer(int ch, int* size);
	void inc_sendbuffer_ptr(int ch, int size);
	uint8* get_recvbuffer0(int ch, int* size0, int* size1);
	uint8* get_recvbuffer1(int ch);
	void inc_recvbuffer_ptr(int ch, int size);
	//@}
	/// @name control menu for user interface
	//@{
	void reset();
	void special_reset();
	void warm_reset();
	void update_config();
	void change_dipswitch(int num);
	void save_state(_TCHAR* filename);
	void load_state(_TCHAR* filename);
	//@}
	/// @name tape menu for user interface
	//@{
	bool play_datarec(_TCHAR* file_path);
	bool rec_datarec(_TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for user interface
	//@{
	bool open_disk(int drv, _TCHAR* file_path, int offset, uint32 flags);
	void close_disk(int drv);
	int  change_disk(int drv);
	bool disk_inserted(int drv);
	int  get_disk_side(int drv);
	//@}
#endif
	/// @name sound menu for user interface
	//@{
	void set_volume();
	//@}
	/// @name printer menu for user interface
	//@{
	void save_printer(int drv, _TCHAR* filename);
	void clear_printer(int drv);
	int  get_printer_buffer_size(int drv);
	uint8* get_printer_buffer(int drv);
	void enable_printer_direct(int drv);
	bool print_printer(int drv);
	//@}
	/// @name comm menu for user interface
	//@{
	void enable_comm_server(int drv);
	void enable_comm_connect(int drv);
//	void recv_comm_data(uint8 *data, size_t size);
	//@}
	/// @name options menu for user interface
	//@{
	void save_keybind();
	//@}

	// ----------------------------------------
	// for each device
	// ----------------------------------------
	/// @name event callbacks
	//@{
	void register_event(DEVICE* device, int event_id, int usec, bool loop, int* register_id);
	void register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id);
	void cancel_event(int register_id);
	void register_frame_event(DEVICE* dev);
	void register_vline_event(DEVICE* dev);
	void set_lines_per_frame(int lines);
	//@}
	/// @name clock
	//@{
	uint64 current_clock();
	uint64 passed_clock(uint64 prev);
	uint32 get_pc();
	//@}
	/// @name get devices
	//@{
	DEVICE* get_device(int id);
	DEVICE* get_device(char *name, char *identifier);
	//@}
};

#endif
