/// @file parsewav.h
///
/// @brief wavef[^ <-> gf[^ <-> VAf[^@ϊNX
///
/// @author Sasaji
/// @date   2011.7.1
///

#ifndef _PARSEWAV_H_
#define _PARSEWAV_H_

#include "../../common.h"
#include "../../fileio.h"
#include "parsewav_param.h"
#include "parsewav_buffer.h"

#pragma pack(1)
typedef struct wav_header_st {
	char RIFF[4];
	uint32 file_len;
	char WAVE[4];
} wav_header_t;
#pragma pack()

#pragma pack(1)
typedef struct wav_fmt_chank_st {
	char fmt[4];
	uint32 fmt_size;
	uint16 format_id;
	uint16 channels;
	uint32 sample_rate;
	uint32 data_speed;
	uint16 block_size;
	uint16 sample_bits;
} wav_fmt_chank_t;
#pragma pack()

#pragma pack(1)
typedef struct wav_data_chank_st {
	char data[4];
	uint32 data_len;
} wav_data_chank_t;
#pragma pack()

#pragma pack(1)
typedef struct wav_unknown_chank_st {
	char data[4];
	uint32 len;
} wav_unknown_chank_t;
#pragma pack()

typedef struct pos_st {
	int    pos;
	uint32 ms;
} pos_t;

class FILEIO;

/// @brief f[^R[_̕ϒg(FSK)͂ăVAf[^ɕϊNX
///
/// wavet@C`ŕۑꂽf[^R[_̕ϒg(FSK)1200Hz2400Hz̔g
/// ͂Ĕgf[^ƂɃVAf[^oB
/// tɃVAf[^ϒgwavet@C쐬B
class PARSEWAV
{
private:
	FILEIO *fio;
	FILEIO *flog;

	int infile_type;

	// input file info
	size_t sample_pos;
	size_t sample_num;
	size_t sample_num_orig;
	uint32 sample_ms;
	int  samples;


	int	phase1;
	int phase2;
	int start_phase;

	int buffer_size;
	int phase2_frip;

	int half_wave;

	struct parse_carrier_st {
		double x0_prev;
		int    wav_prev;
		int    sample_cnt;
		int    odd;
		int    carr_prev;
	} st_pa_carr;

	struct lamda_st {
		double us_delta;	// 1Tv̒(us)
		double us_range;
		double us1200;
		double us1200min;
		double us1200max;
		double us1200avg;
		double us1800;
		double us1800avg;
		double us2400;
		double us2400min;
		double us2400max;
		double us2400avg;
	} st_lamda;

	Buf_t<uint8> *wave_data;
	Buf_t<uint8> *wave_correct_data;
	Buf_t<char> *carrier_data;
	Buf_t<int>   *freq_data;

	int err_num;
	int rep_p1_sample_num[9];
	struct inwav_st {
		wav_header_t	 head;
		wav_fmt_chank_t  fmt;
		wav_data_chank_t data;
	} inwav;

	struct outwav_st {
		wav_header_t	 head;
		wav_fmt_chank_t  fmt;
		wav_data_chank_t data;
	} outwav;

	int remain_buf[100];
	int remain_buf_len;
	int divide;
	int surplus;

	double h[2][50];

	int check_l3c_format(FILEIO *fp);

	int get_wav_sample(int);
	int get_wav_sample(int, int);
	size_t get_wav_sample(Buf_t<uint8> *w_data, int reverse);
	size_t get_l3c_sample(Buf_t<char> *c_data);
	void init_dft(int type);
	size_t calcrate_dft(Buf_t<uint8> *w_data, Buf_t<uint8> *wc_data);
	int   decode_to_carrier_wave_bn(Buf_t<uint8> *w_data, Buf_t<char> *c_data);

	int   find_start_carrier_bit(Buf_t<char> *c_data, int baud, int *frip, char *s_data, char *c_onedata);
	int   decode_to_serial_data(Buf_t<char> *c_data, int baud, int frip, char *s_data, char *c_onedata);
	void  decode_phase1(Buf_t<uint8> *w_data, Buf_t<uint8> *wc_data, Buf_t<char> *c_data, int start_phase, int corret_type, int reverse);
	char  decode_phase2(Buf_t<char> *c_data, char *c_onedata);

	int   encode_phase2(uint8 data, uint8 *buffer);
	int   encode_phase1(uint8 data, uint8 *buffer);

	int   get_sample_rate(int);
	int   out_sample_buffer(int16 *in_buf, size_t in_len, int in_bsiz, uint8 *outbuf, size_t outlen, int outpos, int outbsiz);

public:
	PARSEWAV_PARAM param;

	// prototype define
	PARSEWAV();
	~PARSEWAV();

	int   CheckFileFormat(FILEIO *fp, int file_type);
	int   CheckWavFormat(FILEIO *fp, wav_header_t *head, wav_fmt_chank_t *fmt, wav_data_chank_t *data, size_t *data_len = NULL);

	int   InitData(FILEIO *fp, int file_type);
	uint8 GetData(char *c_onedata = NULL);
	int   SetData(uint8 data, uint8 *buffer);
	int   InitFileHeader(uint8 *buffer);
	void  SetFileHeader(FILEIO *fp);
	void  InitConvSampleData();
	int   OutConvSampleData(void *in_buf, uint32 in_rate, int in_bits, int in_ch, size_t in_len
						, uint8 *outbuf, uint32 outrate, int outbits, size_t outlen, int outpos);
	_TCHAR* Errmsg();
};

#endif
