/** @file volumebox.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2013.10.30 -

	@brief [ volume box ]
*/
#include "volumebox.h"
#include "emu.h"

VolumeBox::VolumeBox(HINSTANCE hInst, EMU *new_emu, config_t *new_cfg)
{
	hInstance = hInst;
	emu       = new_emu;
	pConfig   = new_cfg;
	hVolumeDlg = NULL;
}

VolumeBox::~VolumeBox()
{
	if (hVolumeDlg != NULL) {
		DestroyWindow(hVolumeDlg);
		hVolumeDlg = NULL;
	}
}

HWND VolumeBox::Create(HWND hWnd)
{
	if (hVolumeDlg == NULL) {
		hVolumeDlg = CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_VOLUME), hWnd, VolumeBoxProc, (LPARAM)this);
	}
	return hVolumeDlg;
}

INT_PTR CALLBACK VolumeBox::VolumeBoxProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	VolumeBox *myObj;

	switch (message)
	{
	case WM_INITDIALOG:
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		myObj = (VolumeBox *)lParam;

		return myObj->onInitDialog(hDlg, message, wParam, lParam);

	case WM_HSCROLL:
	case WM_VSCROLL:
		myObj = (VolumeBox *)GetWindowLongPtr(hDlg, DWLP_USER);
		
		myObj->SetVolume(hDlg);

		return (INT_PTR)TRUE;
	case WM_COMMAND:
		myObj = (VolumeBox *)GetWindowLongPtr(hDlg, DWLP_USER);

		myObj->SetVolume(hDlg);

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
		{
			DestroyWindow(hDlg);
			myObj->hVolumeDlg = NULL;

			return (INT_PTR)TRUE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}

INT_PTR VolumeBox::onInitDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL1, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL1, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL1, TBM_SETPOS  , TRUE, 100 - pConfig->volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE1, pConfig->mute);

	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL2, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL2, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL2, TBM_SETPOS  , TRUE, 100 - pConfig->beep_volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE2, pConfig->beep_mute);

	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL3, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL3, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL3, TBM_SETPOS  , TRUE, 100 - pConfig->psg6_volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE3, pConfig->psg6_mute);

	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL4, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL4, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL4, TBM_SETPOS  , TRUE, 100 - pConfig->psg9_volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE4, pConfig->psg9_mute);

	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL5, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL5, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL5, TBM_SETPOS  , TRUE, 100 - pConfig->relay_volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE5, pConfig->relay_mute);

	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL6, TBM_SETRANGE, TRUE, MAKELPARAM(0, 100));
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL6, TBM_SETTICFREQ, 25, 0);
	SendDlgItemMessage(hDlg, IDC_SLIDER_VOL6, TBM_SETPOS  , TRUE, 100 - pConfig->cmt_volume);
	CheckDlgButton(hDlg, IDC_CHK_MUTE6, pConfig->cmt_mute);

	return (INT_PTR)TRUE;
}

void VolumeBox::SetVolume(HWND hDlg)
{
	pConfig->volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL1, TBM_GETPOS, 0, 0);
	pConfig->volume = 100 - pConfig->volume;

	pConfig->beep_volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL2, TBM_GETPOS, 0, 0);
	pConfig->beep_volume = 100 - pConfig->beep_volume;

	pConfig->psg6_volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL3, TBM_GETPOS, 0, 0);
	pConfig->psg6_volume = 100 - pConfig->psg6_volume;

	pConfig->psg9_volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL4, TBM_GETPOS, 0, 0);
	pConfig->psg9_volume = 100 - pConfig->psg9_volume;

	pConfig->relay_volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL5, TBM_GETPOS, 0, 0);
	pConfig->relay_volume = 100 - pConfig->relay_volume;

	pConfig->cmt_volume = (int)SendDlgItemMessage(hDlg, IDC_SLIDER_VOL6, TBM_GETPOS, 0, 0);
	pConfig->cmt_volume = 100 - pConfig->cmt_volume;

	pConfig->mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE1) != 0);
	pConfig->beep_mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE2) != 0);
	pConfig->psg6_mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE3) != 0);
	pConfig->psg9_mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE4) != 0);
	pConfig->relay_mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE5) != 0);
	pConfig->cmt_mute = (IsDlgButtonChecked(hDlg, IDC_CHK_MUTE6) != 0);

	// callback
	if (emu) {
		emu->set_volume(0);
	}

	return;
}
