/** @file winmain.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 main ]
*/

#include <windows.h>
#include <windowsx.h>
#include <winsock.h>
#include <commdlg.h>
#include <commctrl.h>
#include <mmsystem.h>
#include <stdio.h>
#include "winutils.h"
#include "res/resource.h"
#include "config.h"
#include "keybindctrl.h"
#include "emu.h"

#define MIN_INTERVAL	4

#define CALC_NEXT_INTERVAL(now_ms, interval_ms, delta_ms) { \
	interval_ms = now_ms / FRAMES_PER_SEC; \
	now_ms = now_ms - (interval_ms * FRAMES_PER_SEC) + delta_ms; \
}

/// emulation core
EMU* emu;

HINSTANCE hInst;

// buttons
#ifdef USE_BUTTON
#define MAX_FONT_SIZE 32
HFONT hFont[MAX_FONT_SIZE];
HWND hButton[MAX_BUTTONS];
WNDPROC buttonWndProc[MAX_BUTTONS];
#endif

/// key
bool pressed_accelkey = false;
int *key_mod = NULL;

/// menu
bool now_menu = false;
bool now_menuloop = false;

void update_menu(HWND hWnd, HMENU hMenu, int pos);

void show_menu_bar(HWND hWnd)
{
	if(!now_menu) {
		HMENU hMenu = LoadMenu((HINSTANCE)GetModuleHandle(0), MAKEINTRESOURCE(IDR_MENU1));
		SetMenu(hWnd, hMenu);
		now_menu = true;
	}
}

void hide_menu_bar(HWND hWnd)
{
	if(now_menu) {
		HMENU hMenu = GetMenu(hWnd);
		SetMenu(hWnd, NULL);
		DestroyMenu(hMenu);
		now_menu = false;
	}
}

#ifdef _BML3MK5
void execute_global_keys(HWND hWnd, int code, long status);
#endif

static INT_PTR CALLBACK AboutWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

/// screen
//DWORD dwStyle;

/// desktop width on the current monitor
int desktop_width;
/// desktop height on the current monitor
int desktop_height;
/// desktop bpp on the current monitor
int desktop_bpp;

/// client width on the window
int window_client_width;
/// client height on the window
int window_client_height;

int min_window_width;
int min_window_height;
int max_window_width;
int max_window_height;

/// timing control
#define MAX_SKIP_FRAMES 10

int fskip_remain = 0;
static const int fps[6] = {60, 30, 20, 15, 12, 10};
static const int fskip[6] = {1, 2, 3, 4, 5, 6};
int rec_fps_no = -1;

/*
int get_interval()
{
	static int accum = 0;
	accum += emu->frame_interval();
	int interval = accum >> 10;
	accum -= interval << 10;
	return interval;
}
*/

/// windows main
int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR szCmdLine, int iCmdShow);
LRESULT CALLBACK WndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
#ifdef USE_BUTTON
LRESULT CALLBACK ButtonWndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
#endif

int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR szCmdLine, int iCmdShow)
{
	// initialize d88 file info
	memset(d88_file, 0, sizeof(d88_file));

	InitCommonControls();
	CoInitialize(NULL);

	emu = new EMU();

	// load config
	load_config();

	hInst = hInstance;

	// create window
	WNDCLASSEX wndclass;
	wndclass.cbSize = sizeof(WNDCLASSEX);
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = (WNDPROC)WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
	wndclass.lpszClassName = _T(CLASS_NAME);
	wndclass.hIconSm = NULL;
	RegisterClassEx(&wndclass);

#ifdef _BML3MK5
	KeyBindControl_RegisterClass(hInstance);
#endif
	// get window client size
	//  | WS_THICKFRAME
	DWORD dwStyle =  WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE;

	RECT rect;
	SetRect(&rect, 0, 0, MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT);
	AdjustWindowRectEx(&rect, dwStyle, TRUE, 0);
	min_window_width = (rect.right - rect.left);
	min_window_height = (rect.bottom - rect.top);
	SetRect(&rect, 0, 0, MAX_WINDOW_WIDTH, MAX_WINDOW_HEIGHT);
	AdjustWindowRectEx(&rect, dwStyle, TRUE, 0);
	max_window_width = (rect.right - rect.left);
	max_window_height = (rect.bottom - rect.top);
	// get window position
	window_client_width = MIN_WINDOW_WIDTH;
	window_client_height = MIN_WINDOW_HEIGHT;
	SetRect(&rect, 0, 0, window_client_width, window_client_height);
	AdjustWindowRectEx(&rect, dwStyle, TRUE, 0);
	HDC hdcScr = GetDC(NULL);
	desktop_width = GetDeviceCaps(hdcScr, HORZRES);
	desktop_height = GetDeviceCaps(hdcScr, VERTRES);
	desktop_bpp = GetDeviceCaps(hdcScr, BITSPIXEL);
	ReleaseDC(NULL, hdcScr);
	int dest_x = (int)((desktop_width - (rect.right - rect.left)) / 2);
	int dest_y = (int)((desktop_height - (rect.bottom - rect.top)) / 2);
	//dest_x = (dest_x < 0) ? 0 : dest_x;
	dest_y = (dest_y < 0) ? 0 : dest_y;

	// show window
	//  | WS_THICKFRAME
	HWND hWnd = CreateWindow(_T(CLASS_NAME), _T(DEVICE_NAME), dwStyle,
	                         dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, NULL, NULL, hInstance, NULL);

	ShowWindow(hWnd, iCmdShow);
	UpdateWindow(hWnd);

	// show menu
	show_menu_bar(hWnd);

	// enumerate screen mode
	emu->set_window_style(dwStyle);
	emu->set_desktop_size(desktop_width, desktop_height, desktop_bpp);
	emu->enum_window_mode(desktop_width, desktop_height);
	emu->enum_screen_mode();
	emu->find_screen_mode();

	// restore screen mode
	if(config.window_mode >= 0 && config.window_mode < 8) {
		PostMessage(hWnd, WM_COMMAND, ID_SCREEN_WINDOW1 + config.window_mode, 0L);
	}
	else if(config.window_mode >= 8) {
		PostMessage(hWnd, WM_COMMAND, ID_SCREEN_FULLSCREEN1 + config.window_mode - 8, 0L);
	}

	// accelerator
	HACCEL hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR1));

	// disable ime
	ImmAssociateContext(hWnd, NULL);

	// initialize emulation core
	emu->initialize(hWnd, hInstance);
//	emu->set_display_size(window_client_width, window_client_height, 1, true);
	key_mod = emu->get_key_mod_ptr();

	// open command line path
	if(szCmdLine[0]) {
		if(szCmdLine[0] == _T('"')) {
			int len = (int)_tcslen(szCmdLine);
			szCmdLine[len - 1] = _T('\0');
			szCmdLine++;
		}
		_TCHAR path[_MAX_PATH];
		get_long_full_path_name(szCmdLine, path);
#ifdef USE_CART
		UPDATE_HISTORY(path, config.recent_cart_path);
		emu->open_cart(path);
		_tcscpy(config.initial_cart_path, get_parent_dir(path));
#elif defined(USE_FD1)
		UPDATE_HISTORY(path, config.recent_disk_path[0]);
		if (open_disk(0, path, 0, 0)) {
			inserted_disk_drive |= 1;
		} else {
			inserted_disk_drive &= ~1;
		}
		_tcscpy(config.initial_disk_path, get_parent_dir(path));
#elif defined(USE_QUICKDISK)
		UPDATE_HISTORY(path, config.recent_quickdisk_path);
		emu->open_quickdisk(path);
		_tcscpy(config.initial_quickdisk_path, get_parent_dir(path));
#endif
	} else {
		// auto open recent file
#ifdef USE_FD1
		_TCHAR path[_MAX_PATH];
		for(int i=0; i<MAX_DRIVE; i++) {
			if (_tcslen(config.recent_disk_path[i][0]) > 0)
				_tcscpy(path, config.recent_disk_path[i][0]);
			if (open_disk(i, path, 0, 0)) {
				inserted_disk_drive |= (1 << i);
			} else {
				inserted_disk_drive &= ~(1 << i);
			}
		}
//		_tcscpy(config.initial_disk_path, get_parent_dir(path));
#endif
	}

	// vm start
	if (emu) {
		emu->reset();
	}

	// main loop
	int total_frames = 0, draw_frames = 0, skip_frames = 0;
	int rec_delay_ptr = 0;

	int ms = 1000;
	int next_interval;
	CALC_NEXT_INTERVAL(ms, next_interval, 1000);

	int split_num = 0;

	DWORD start_time = 0;
	DWORD current_time = 0;
	int   remain_time = 0;
	DWORD next_time = 0;
	DWORD update_fps_time = next_time + 1000;
	MSG msg;
	int msgcnt = 0;
	_TCHAR buf[256];

	timeBeginPeriod(MIN_INTERVAL);

	int return_code = 0;
	bool working = true;
	// process coming messages at first. 
	while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
		if(!GetMessage(&msg, NULL, 0, 0)) {
//			ExitProcess(0);	// trick
			return_code = (int)msg.wParam;
			working = false;
			break;
		}
		if(!TranslateAccelerator(hWnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	next_time = timeGetTime();
	// loop
	while(working) {
		split_num = 0;
		// get next period
		CALC_NEXT_INTERVAL(ms, next_interval, 1000);

		start_time = next_time;
//		next_time += emu->now_skip() ? 0 : next_interval;
		next_time += next_interval;
		current_time = timeGetTime();

		// when vm processing is too late...
		if (next_time + 200 <= current_time) {
			// adjust next time for current time
//			next_time = current_time + (emu->now_skip() ? 0 : next_interval);
			next_time = current_time + next_interval;
		}
//		emu->out_log(LOG_DEBUG, "A1: cur:%ld st:%ld nt:%ld ms:%d int:%d", current_time, start_time, next_time, ms, next_interval);

		for(int i=0; i < (FRAME_SPLIT_NUM - 1); i++) {
			// drive machine
			if(emu)	emu->run(split_num);

			msgcnt = 0;
			while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) && msgcnt < 10) {
				if(!GetMessage(&msg, NULL, 0, 0)) {
//					ExitProcess(0);	// trick
					return_code = (int)msg.wParam;
					working = false;
					goto FIN;
					break;
				}
				if(!TranslateAccelerator(hWnd, hAccel, &msg)) {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
				msgcnt++;
			}
			split_num++;
		}

		if(emu) {
			// drive machine
			emu->run(split_num);

//			if (split_num + 1 == FRAME_SPLIT_NUM) {
				total_frames++;
				if(emu->now_rec_video()) {
					if (fskip_remain <= 0) {
						// rec pictures constant frames per 1 second
						emu->mix_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[rec_fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else if(config.fps_no >= 0) {
					if (fskip_remain <= 0) {
						// constant frames per 1 second
						emu->mix_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[config.fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else {
					current_time = timeGetTime();
					if(next_time > current_time) {
						// update window if enough time
						emu->mix_screen();
						draw_frames++;
						skip_frames = 0;
					}
					else if(++skip_frames > MAX_SKIP_FRAMES) {
						// update window at least once per 10 frames
						emu->mix_screen();
						draw_frames++;
						skip_frames = 0;
						next_time = timeGetTime();
					}
					else {
						emu->skip_screen();
					}
				}
//			}
		}
//		emu->out_log(LOG_DEBUG,"A2: rti:%ld nt:%ld",timeGetTime(),next_time);
		// process event at left time in 1 frame if need
		do {
//			current_time = timeGetTime();
//			remain_time = next_time - current_time;
			// check window message
			msgcnt = 0;
			while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) && msgcnt < 100) {
				if(!GetMessage(&msg, NULL, 0, 0)) {
//					ExitProcess(0);	// trick
					return_code = (int)msg.wParam;
					working = false;
					goto FIN;
					break;
				}
				if(!TranslateAccelerator(hWnd, hAccel, &msg)) {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
				msgcnt++;
			}
			if (skip_frames == 0) {
				current_time = timeGetTime();
				remain_time = next_time - current_time;
				if (remain_time > MIN_INTERVAL) {
					if (emu) emu->update();
					current_time = timeGetTime();
					remain_time = next_time - current_time;
					if (remain_time > MIN_INTERVAL) {
						Sleep(MIN_INTERVAL);
						current_time = timeGetTime();
						remain_time = next_time - current_time;
					}
				}
			} else {
				remain_time = 0;
			}
		} while (remain_time > 1);

//		split_num = (split_num + 1) & (FRAME_SPLIT_NUM - 1);

//		emu->out_debug("A2: rti:%ld nt:%ld (total:%d fps:%d skip:%d)",timeGetTime(),next_time,total_frames,draw_frames,skip_frames);
//		emu->out_log(LOG_DEBUG,"A3: rti:%ld nt:%ld",timeGetTime(),next_time);
//		Sleep(0);

		// calc frame rate
		current_time = timeGetTime();
		if(update_fps_time <= current_time) {
			int ratio = 0;
			if (total_frames) ratio = (int)(100.0 * (double)draw_frames / (double)total_frames + 0.5);
			_stprintf(buf, _T("%s - %d / %d fps (%d %%)"), _T(DEVICE_NAME), draw_frames, total_frames, ratio);
			SetWindowText(hWnd, buf);

			update_fps_time += 1000;
			if(update_fps_time <= current_time) {
				update_fps_time = current_time + 1000;
			}
			total_frames = draw_frames = 0;
		}
	}

FIN:
	timeEndPeriod(MIN_INTERVAL);

	// exit process
#ifdef _DEBUG
	_CrtDumpMemoryLeaks();
#endif
	return return_code;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	_TCHAR path[_MAX_PATH];
	int no;

	switch(iMsg) {
	case WM_CREATE:
#ifdef USE_BUTTON
		memset(hFont, 0, sizeof(hFont));
		for(int i = 0; i < MAX_BUTTONS; i++) {
			hButton[i] = CreateWindow(_T("BUTTON"), buttons[i].caption,
			                          WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_MULTILINE,
			                          buttons[i].x, buttons[i].y,
			                          buttons[i].width, buttons[i].height,
			                          hWnd, (HMENU)(ID_BUTTON + i), (HINSTANCE)GetModuleHandle(0), NULL);
			buttonWndProc[i] = (WNDPROC)(LONG_PTR)GetWindowLong(hButton[i], GWL_WNDPROC);
			SetWindowLong(hButton[i], GWL_WNDPROC, (LONG)(LONG_PTR)ButtonWndProc);
			//HFONT hFont = GetWindowFont(hButton[i]);
			if(!hFont[buttons[i].font_size]) {
				LOGFONT logfont;
				logfont.lfEscapement = 0;
				logfont.lfOrientation = 0;
				logfont.lfWeight = FW_NORMAL;
				logfont.lfItalic = FALSE;
				logfont.lfUnderline = FALSE;
				logfont.lfStrikeOut = FALSE;
				logfont.lfCharSet = DEFAULT_CHARSET;
				logfont.lfOutPrecision = OUT_TT_PRECIS;
				logfont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
				logfont.lfQuality = DEFAULT_QUALITY;
				logfont.lfPitchAndFamily = FIXED_PITCH | FF_DONTCARE;
				_tcscpy(logfont.lfFaceName, _T("Arial"));
				logfont.lfHeight = buttons[i].font_size;
				logfont.lfWidth = buttons[i].font_size >> 1;
				hFont[buttons[i].font_size] = CreateFontIndirect(&logfont);
			}
			SetWindowFont(hButton[i], hFont[buttons[i].font_size], TRUE);
		}
#endif
		break;
	case WM_CLOSE:
		if(emu) {
#ifdef USE_POWER_OFF
			// notify power off
			static int notified = 0;
			if(!notified) {
				emu->notify_power_off();
				notified = 1;
				return 0;
			}
#endif
		}
		// quit fullscreen mode
		if(emu->is_fullscreen()) {
			ChangeDisplaySettings(NULL, 0);
		}
//		emu->is_fullscreen() = false;
#ifdef USE_BUTTON
		for(int i = 0; i < MAX_FONT_SIZE; i++) {
			if(hFont[i]) {
				DeleteObject(hFont[i]);
			}
		}
#endif
		// release emulation core
		if(emu) {
			emu->release();
		}

		save_config();
		release_config();

		if(emu) {
			delete emu;
		}
		emu = NULL;
		DestroyWindow(hWnd);
		CoUninitialize();
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
#ifdef USE_BITMAP
	case WM_SIZE:
		if(emu) {
			emu->reload_bitmap();
		}
		break;
#endif
	case WM_KILLFOCUS:
		if(emu) {
			emu->key_lost_focus();
		}
		break;
	case WM_PAINT:
		if(emu) {
			PAINTSTRUCT ps;
			HDC hdc;
//			emu->out_debug("P1: %ld",timeGetTime());
			hdc = BeginPaint(hWnd, &ps);
//			emu->out_debug("P2: %ld",timeGetTime());
			emu->update_screen(hdc);
			EndPaint(hWnd, &ps);
//			emu->out_debug("P3: %ld",timeGetTime());
		}
		return 0;
	case WM_ENTERSIZEMOVE:
		if(emu) {
//			emu->mute_sound(true);
			emu->set_pause(1, true);
		}
		break;
	case WM_EXITSIZEMOVE:
		if(emu) {
			emu->key_lost_focus();
//			emu->mute_sound(false);
			emu->set_pause(1, false);
		}
		break;
	case WM_MOVE:
		if(emu) {
			emu->move_led();
		}
		break;
	case WM_MOVING:
		if(emu) {
//			emu->mute_sound(true);
			emu->set_pause(1, true);
		}
		break;
	case WM_KEYDOWN:
		if ((*key_mod) & KEY_MOD_ALT_KEY) {
			execute_global_keys(hWnd, LOBYTE(wParam), (long)lParam);
			return 0;
		} else if(emu) {
//			bool repeat = ((HIWORD(lParam) & 0x4000) != 0);
//			emu->key_down(LOBYTE(wParam), repeat);
			emu->key_down(LOBYTE(wParam), (long)lParam);
		}
		break;
	case WM_KEYUP:
		pressed_accelkey = false;
		if(emu) {
//			emu->key_up(LOBYTE(wParam));
			emu->key_up(LOBYTE(wParam), (long)lParam);
		}
		break;
	case WM_SYSKEYDOWN:
		if ((*key_mod) & KEY_MOD_ALT_KEY) {
			execute_global_keys(hWnd, LOBYTE(wParam), (long)lParam);
			return 0;
		} else if(emu) {
//			bool repeat = ((HIWORD(lParam) & 0x4000) != 0);
//			emu->key_down(LOBYTE(wParam), repeat);
			emu->key_down(LOBYTE(wParam), (long)lParam);
		}
#ifdef USE_ALT_F10_KEY
		return 0;	// not activate menu when hit ALT/F10
#endif
		break;
	case WM_SYSKEYUP:
		pressed_accelkey = false;
		if(emu) {
//			emu->key_up(LOBYTE(wParam));
			emu->key_up(LOBYTE(wParam), (long)lParam);
		}
#ifdef USE_ALT_F10_KEY
		return 0;	// not activate menu when hit ALT/F10
#endif
		break;
	case WM_SYSCHAR:
		return 0;	// not activate menu
		break;
	case WM_INITMENUPOPUP:
//		if(emu) {
//			emu->mute_sound(true);
//			emu->set_pause(1, true);
//		}
		update_menu(hWnd, (HMENU)wParam, LOWORD(lParam));
		break;
	case WM_ENTERMENULOOP:
		if(emu) {
			emu->set_pause(1, true);
		}
		now_menuloop = true;
		break;
	case WM_EXITMENULOOP:
		if(emu->is_fullscreen() && now_menuloop) {
			hide_menu_bar(hWnd);
		}
		if (emu) {
			emu->key_lost_focus();
//			emu->mute_sound(false);
			emu->set_pause(1, false);
		}
		now_menuloop = false;
		break;
	case WM_MOUSEMOVE:
		if(emu->is_fullscreen() && !now_menuloop) {
			POINTS p = MAKEPOINTS(lParam);
			if(p.y == 0) {
				show_menu_bar(hWnd);
			}
			else if(p.y > 32) {
				hide_menu_bar(hWnd);
			}
		}
		break;
	case WM_SIZING:
		{
			RECT *re = (RECT*)lParam;
			int x = re->left;
			int y = re->top;
			int w = re->right - x;
			int h = re->bottom - y;
			if (w < min_window_width) {
				re->right = x + min_window_width;
			}
			if (max_window_width < w) {
				re->right = x + max_window_width;
			}
			if (h < min_window_height) {
				re->bottom = y + min_window_height;
			}
			if (max_window_height < h) {
				re->bottom = y + max_window_height;
			}
		}
		break;
	case WM_SIZE:
		{
			window_client_width = LOWORD(lParam);
			window_client_height = HIWORD(lParam);
//			if (emu) {
//				emu->set_display_size(window_client_width, window_client_height, window_mode_power, true);
//			}
		}
		break;
	case WM_RESIZE:
		if(emu) {
//			if(emu->is_fullscreen()) {
//				emu->set_display_size(-1, -1, 10, false);
//			}
//			else {
//				emu->set_window(hWnd, config.window_mode);
//			}
			emu->change_screen_mode(hWnd, config.window_mode);
		}
		break;
#ifdef USE_SOCKET
	case WM_SOCKET0:
	case WM_SOCKET1:
	case WM_SOCKET2:
	case WM_SOCKET3:
	case WM_SOCKET4:
	case WM_SOCKET5:
		no = iMsg - WM_SOCKET0;
		if(!emu) {
			break;
		}
		if(WSAGETSELECTERROR(lParam) != 0) {
			emu->disconnect_socket(no);
			emu->socket_disconnected(no);
			break;
		}
		if(emu->get_socket(no) != (int)wParam) {
			break;
		}
		switch(WSAGETSELECTEVENT(lParam)) {
		case FD_CONNECT:
			emu->socket_connected(no);
			break;
		case FD_CLOSE:
			emu->socket_disconnected(no);
			break;
		case FD_WRITE:
			emu->socket_writeable(no);
			emu->send_data(no);
			break;
		case FD_READ:
			emu->recv_data(no);
			emu->socket_readable(no);
			break;
		case FD_ACCEPT:
			emu->socket_accept(no);
			break;
		}
		break;
#endif
#ifdef _BML3MK5
#if 0
	case WM_COPYDATA:
		return copy_data(hWnd, lParam);
		break;
#endif
#endif
	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case ID_RESET:
			if (!pressed_accelkey) {
				if (HIWORD(wParam)) pressed_accelkey = true;
				if(emu) {
					emu->reset();
				}
//				rec_next_time = rec_accum_time = 0;
			}
			break;
#ifdef USE_SPECIAL_RESET
		case ID_SPECIAL_RESET:
			if (!pressed_accelkey) {
				if (HIWORD(wParam)) pressed_accelkey = true;
				if(emu) {
					emu->special_reset();
				}
//				rec_next_time = rec_accum_time = 0;
			}
			break;
		case ID_WARM_RESET:
			if (!pressed_accelkey) {
				if (HIWORD(wParam)) pressed_accelkey = true;
				if(emu) {
					emu->warm_reset();
				}
//				rec_next_time = rec_accum_time = 0;
			}
			break;
#endif
#ifdef USE_DIPSWITCH
		case ID_DIPSWITCH1:
		case ID_DIPSWITCH2:
		case ID_DIPSWITCH3:
		case ID_DIPSWITCH4:
		case ID_DIPSWITCH5:
		case ID_DIPSWITCH6:
		case ID_DIPSWITCH7:
		case ID_DIPSWITCH8:
			if(emu) {
				emu->change_dipswitch(LOWORD(wParam) - ID_DIPSWITCH1);
			}
			break;
#endif
#ifdef _HC80
		case ID_HC80_RAMDISK0:
		case ID_HC80_RAMDISK1:
		case ID_HC80_RAMDISK2:
			config.device_type = LOWORD(wParam) - ID_HC80_RAMDISK0;
			break;
#endif
#ifdef _MZ800
		case ID_MZ800_MODE_MZ800:
		case ID_MZ800_MODE_MZ700:
			config.boot_mode = LOWORD(wParam) - ID_MZ800_MODE_MZ800;
			if(emu) {
				emu->update_config();
			}
			break;
#endif
#ifdef _PASOPIA
		case ID_PASOPIA_MODE_TBASIC_V1_0:
		case ID_PASOPIA_MODE_TBASIC_V1_1:
		case ID_PASOPIA_MODE_OABASIC:
		case ID_PASOPIA_MODE_OABASIC_NO_DISK:
		case ID_PASOPIA_MODE_MINI_PASCAL:
			config.boot_mode = LOWORD(wParam) - ID_PASOPIA_MODE_TBASIC_V1_0;
			if(emu) {
				emu->update_config();
			}
			break;
		case ID_PASOPIA_DEVICE_RAM_PAC:
		case ID_PASOPIA_DEVICE_KANJI_ROM:
		case ID_PASOPIA_DEVICE_JOYSTICK:
			config.device_type = LOWORD(wParam) - ID_PASOPIA_DEVICE_RAM_PAC;
			break;
#endif
#ifdef _PC98DO
		case ID_PC98DO_MODE_PC98:
		case ID_PC8801_MODE_V1S:
		case ID_PC8801_MODE_V1H:
		case ID_PC8801_MODE_V2:
		case ID_PC8801_MODE_N:
			config.boot_mode = LOWORD(wParam) - ID_PC98DO_MODE_PC98;
			if(emu) {
				emu->update_config();
			}
			break;
#endif
#ifdef _PC8801MA
		case ID_PC8801_MODE_V1S:
		case ID_PC8801_MODE_V1H:
		case ID_PC8801_MODE_V2:
		case ID_PC8801_MODE_N:
			config.boot_mode = LOWORD(wParam) - ID_PC8801_MODE_V1S;
			if(emu) {
				emu->update_config();
			}
			break;
		case ID_PC8801_DEVICE_JOYSTICK:
		case ID_PC8801_DEVICE_MOUSE:
		case ID_PC8801_DEVICE_JOYMOUSE:
			config.device_type = LOWORD(wParam) - ID_PC8801_DEVICE_JOYSTICK;
			break;
#endif
#if defined(_PC9801E) || defined(_PC9801VM) || defined(_PC98DO) || defined(_PC8801MA)
		case ID_PC9801_CPU_CLOCK_HIGH:
			config.cpu_clock_low = false;
			break;
		case ID_PC9801_CPU_CLOCK_LOW:
			config.cpu_clock_low = true;
			break;
#endif
		case ID_CPU_POWER0:
		case ID_CPU_POWER1:
		case ID_CPU_POWER2:
		case ID_CPU_POWER3:
		case ID_CPU_POWER4:
		case ID_CPU_POWER5:
			if(emu) {
				emu->change_cpu_power(LOWORD(wParam) - ID_CPU_POWER0);
			}
			break;
#ifdef _BML3MK5
		case ID_PAUSE:
			if(emu) {
				emu->set_pause(2, emu->get_pause(2) ? false : true);
			}
			break;
		case ID_SYNC_IRQ:
			if(emu) {
				emu->change_sync_irq();
			}
			break;
#endif
#ifdef USE_AUTO_KEY
		case ID_AUTOKEY_START:
			if(emu) {
				emu->start_auto_key();
			}
			break;
		case ID_AUTOKEY_OPEN:
			open_autokey_dialog(hWnd);
			break;
		case ID_AUTOKEY_STOP:
			if(emu) {
				emu->stop_auto_key();
			}
			break;
#endif
		case ID_LOAD_STATE:
			open_load_state_dialog(hWnd);
			break;
		case ID_SAVE_STATE:
			open_save_state_dialog(hWnd);
			break;
		case ID_EXIT:
			SendMessage(hWnd, WM_CLOSE, 0, 0L);
			break;
#ifdef USE_CART
		case ID_OPEN_CART:
			if(emu) {
				open_cart_dialog(hWnd);
			}
			break;
		case ID_CLOSE_CART:
			if(emu) {
				emu->close_cart();
			}
			break;
		case ID_RECENT_CART + 0:
		case ID_RECENT_CART + 1:
		case ID_RECENT_CART + 2:
		case ID_RECENT_CART + 3:
		case ID_RECENT_CART + 4:
		case ID_RECENT_CART + 5:
		case ID_RECENT_CART + 6:
		case ID_RECENT_CART + 7:
			no = LOWORD(wParam) - ID_RECENT_CART;
			_tcscpy(path, config.recent_cart_path[no]);
			for(int i = no; i > 0; i--) {
				_tcscpy(config.recent_cart_path[i], config.recent_cart_path[i - 1]);
			}
			_tcscpy(config.recent_cart_path[0], path);
			if(emu) {
				emu->open_cart(path);
			}
			break;
#endif
#ifdef USE_FD1
		#define FD_MENU_ITEMS(drv, ID_OPEN_FD, ID_CLOSE_FD, ID_CHANGE_FD, ID_RECENT_FD, ID_SELECT_D88_BANK) \
		case ID_OPEN_FD: \
			if(emu) { \
				open_disk_dialog(hWnd, drv); \
			} \
			break; \
		case ID_CLOSE_FD: \
			if(emu) { \
				close_disk(drv); \
			} \
			break; \
		case ID_CHANGE_FD: \
			if(emu) { \
				emu->change_disk(drv); \
			} \
			break; \
		case ID_RECENT_FD + 0: \
		case ID_RECENT_FD + 1: \
		case ID_RECENT_FD + 2: \
		case ID_RECENT_FD + 3: \
		case ID_RECENT_FD + 4: \
		case ID_RECENT_FD + 5: \
		case ID_RECENT_FD + 6: \
		case ID_RECENT_FD + 7: \
			no = LOWORD(wParam) - ID_RECENT_FD; \
			_tcscpy(path, config.recent_disk_path[drv][no]); \
			for(int i = no; i > 0; i--) { \
				_tcscpy(config.recent_disk_path[drv][i], config.recent_disk_path[drv][i - 1]); \
			} \
			_tcscpy(config.recent_disk_path[drv][0], path); \
			if(emu) { \
				if (emu->open_disk(drv, path, 0)) { \
					inserted_disk_drive |= (1 << drv); \
					_tcscpy(config.initial_disk_path, get_parent_dir(path)); \
				} else { \
					inserted_disk_drive &= ~(1 << drv); \
					emu->out_log(LOG_WARN, _T("Floppy image couldn't be opend.")); \
				} \
			} \
			break; \
		case ID_SELECT_D88_BANK +  0: \
		case ID_SELECT_D88_BANK +  1: \
		case ID_SELECT_D88_BANK +  2: \
		case ID_SELECT_D88_BANK +  3: \
		case ID_SELECT_D88_BANK +  4: \
		case ID_SELECT_D88_BANK +  5: \
		case ID_SELECT_D88_BANK +  6: \
		case ID_SELECT_D88_BANK +  7: \
		case ID_SELECT_D88_BANK +  8: \
		case ID_SELECT_D88_BANK +  9: \
		case ID_SELECT_D88_BANK + 10: \
		case ID_SELECT_D88_BANK + 11: \
		case ID_SELECT_D88_BANK + 12: \
		case ID_SELECT_D88_BANK + 13: \
		case ID_SELECT_D88_BANK + 14: \
		case ID_SELECT_D88_BANK + 15: \
		case ID_SELECT_D88_BANK + 16: \
		case ID_SELECT_D88_BANK + 17: \
		case ID_SELECT_D88_BANK + 18: \
		case ID_SELECT_D88_BANK + 19: \
		case ID_SELECT_D88_BANK + 20: \
		case ID_SELECT_D88_BANK + 21: \
		case ID_SELECT_D88_BANK + 22: \
		case ID_SELECT_D88_BANK + 23: \
		case ID_SELECT_D88_BANK + 24: \
		case ID_SELECT_D88_BANK + 25: \
		case ID_SELECT_D88_BANK + 26: \
		case ID_SELECT_D88_BANK + 27: \
		case ID_SELECT_D88_BANK + 28: \
		case ID_SELECT_D88_BANK + 29: \
		case ID_SELECT_D88_BANK + 30: \
		case ID_SELECT_D88_BANK + 31: \
		case ID_SELECT_D88_BANK + 32: \
		case ID_SELECT_D88_BANK + 33: \
		case ID_SELECT_D88_BANK + 34: \
		case ID_SELECT_D88_BANK + 35: \
		case ID_SELECT_D88_BANK + 36: \
		case ID_SELECT_D88_BANK + 37: \
		case ID_SELECT_D88_BANK + 38: \
		case ID_SELECT_D88_BANK + 39: \
		case ID_SELECT_D88_BANK + 40: \
		case ID_SELECT_D88_BANK + 41: \
		case ID_SELECT_D88_BANK + 42: \
		case ID_SELECT_D88_BANK + 43: \
		case ID_SELECT_D88_BANK + 44: \
		case ID_SELECT_D88_BANK + 45: \
		case ID_SELECT_D88_BANK + 46: \
		case ID_SELECT_D88_BANK + 47: \
		case ID_SELECT_D88_BANK + 48: \
		case ID_SELECT_D88_BANK + 49: \
			no = LOWORD(wParam) - ID_SELECT_D88_BANK; \
			if(emu && d88_file[drv].cur_bank != no) { \
				if (emu->open_disk(drv, d88_file[drv].path, d88_file[drv].bank[no].offset)) { \
					inserted_disk_drive |= (1 << drv); \
					d88_file[drv].cur_bank = no; \
				} else { \
					inserted_disk_drive &= ~(1 << drv); \
				} \
			} \
			break;
		FD_MENU_ITEMS(0, ID_OPEN_FD1, ID_CLOSE_FD1, ID_CHANGE_FD1, ID_RECENT_FD1, ID_SELECT_D88_BANK1)
#endif
#ifdef USE_FD2
		FD_MENU_ITEMS(1, ID_OPEN_FD2, ID_CLOSE_FD2, ID_CHANGE_FD2, ID_RECENT_FD2, ID_SELECT_D88_BANK2)
#endif
#ifdef USE_FD3
		FD_MENU_ITEMS(2, ID_OPEN_FD3, ID_CLOSE_FD3, ID_CHANGE_FD3, ID_RECENT_FD3, ID_SELECT_D88_BANK3)
#endif
#ifdef USE_FD4
		FD_MENU_ITEMS(3, ID_OPEN_FD4, ID_CLOSE_FD4, ID_CHANGE_FD4, ID_RECENT_FD4, ID_SELECT_D88_BANK4)
#endif
#ifdef USE_FD5
		FD_MENU_ITEMS(4, ID_OPEN_FD5, ID_CLOSE_FD5, ID_CHANGE_FD5, ID_RECENT_FD5, ID_SELECT_D88_BANK5)
#endif
#ifdef USE_FD6
		FD_MENU_ITEMS(5, ID_OPEN_FD6, ID_CLOSE_FD6, ID_CHANGE_FD6, ID_RECENT_FD6, ID_SELECT_D88_BANK6)
#endif
#ifdef USE_QUICKDISK
		case ID_OPEN_QUICKDISK:
			if(emu) {
				open_quickdisk_dialog(hWnd);
			}
			break;
		case ID_CLOSE_QUICKDISK:
			if(emu) {
				emu->close_quickdisk();
			}
			break;
		case ID_RECENT_QUICKDISK + 0:
		case ID_RECENT_QUICKDISK + 1:
		case ID_RECENT_QUICKDISK + 2:
		case ID_RECENT_QUICKDISK + 3:
		case ID_RECENT_QUICKDISK + 4:
		case ID_RECENT_QUICKDISK + 5:
		case ID_RECENT_QUICKDISK + 6:
		case ID_RECENT_QUICKDISK + 7:
			no = LOWORD(wParam) - ID_RECENT_QUICKDISK;
			_tcscpy(path, config.recent_quickdisk_path[no]);
			for(int i = no; i > 0; i--) {
				_tcscpy(config.recent_quickdisk_path[i], config.recent_quickdisk_path[i - 1]);
			}
			_tcscpy(config.recent_quickdisk_path[0], path);
			if(emu) {
				emu->open_quickdisk(path);
			}
			break;
#endif
#ifdef USE_DATAREC
		case ID_PLAY_DATAREC:
			if(emu) {
				open_datarec_dialog(hWnd, true);
			}
			break;
		case ID_REC_DATAREC:
			if(emu) {
				open_datarec_dialog(hWnd, false);
			}
			break;
		case ID_CLOSE_DATAREC:
			if(emu) {
				opened_tape_type = 0;
				emu->close_datarec();
			}
			break;
		case ID_REWIND_DATAREC:
			if(emu) {
				emu->rewind_datarec();
			}
			break;
		case ID_FAST_FORWARD_DATAREC:
			if(emu) {
				emu->fast_forward_datarec();
			}
			break;
		case ID_REAL_DATAREC:
			if(emu) {
				emu->realmode_datarec();
			}
			break;
		case ID_RECENT_DATAREC + 0:
		case ID_RECENT_DATAREC + 1:
		case ID_RECENT_DATAREC + 2:
		case ID_RECENT_DATAREC + 3:
		case ID_RECENT_DATAREC + 4:
		case ID_RECENT_DATAREC + 5:
		case ID_RECENT_DATAREC + 6:
		case ID_RECENT_DATAREC + 7:
			no = LOWORD(wParam) - ID_RECENT_DATAREC;
			_tcscpy(path, config.recent_datarec_path[no]);
			for(int i = no; i > 0; i--) {
				_tcscpy(config.recent_datarec_path[i], config.recent_datarec_path[i - 1]);
			}
			_tcscpy(config.recent_datarec_path[0], path);
			if(emu) {
				if (emu->play_datarec(path)) {
					opened_tape_type = 1;
					_tcscpy(config.initial_datarec_path, get_parent_dir(path));
				} else {
					opened_tape_type = 0;
					emu->out_log(LOG_WARN, _T("Tape image couldn't be opend."));
				}
			}
			break;
#endif
#ifdef USE_DATAREC_BUTTON
		case ID_PLAY_BUTTON:
			if(emu) {
				emu->push_play();
			}
			break;
		case ID_STOP_BUTTON:
			if(emu) {
				emu->push_stop();
			}
			break;
#endif
#ifdef USE_MEDIA
		case ID_OPEN_MEDIA:
			if(emu) {
				open_media_dialog(hWnd);
			}
			break;
		case ID_CLOSE_MEDIA:
			if(emu) {
				emu->close_media();
			}
			break;
		case ID_RECENT_MEDIA + 0:
		case ID_RECENT_MEDIA + 1:
		case ID_RECENT_MEDIA + 2:
		case ID_RECENT_MEDIA + 3:
		case ID_RECENT_MEDIA + 4:
		case ID_RECENT_MEDIA + 5:
		case ID_RECENT_MEDIA + 6:
		case ID_RECENT_MEDIA + 7:
			no = LOWORD(wParam) - ID_RECENT_MEDIA;
			_tcscpy(path, config.recent_media_path[no]);
			for(int i = no; i > 0; i--) {
				_tcscpy(config.recent_media_path[i], config.recent_media_path[i - 1]);
			}
			_tcscpy(config.recent_media_path[0], path);
			if(emu) {
				emu->open_media(path);
			}
			break;
#endif
#ifdef USE_BINARY_FILE1
		#define BINARY_MENU_ITEMS(drv, ID_LOAD_BINARY, ID_SAVE_BINARY, ID_RECENT_BINARY) \
		case ID_LOAD_BINARY: \
			if(emu) { \
				open_binary_dialog(hWnd, drv, true); \
			} \
			break; \
		case ID_SAVE_BINARY: \
			if(emu) { \
				open_binary_dialog(hWnd, drv, false); \
			} \
			break; \
		case ID_RECENT_BINARY + 0: \
		case ID_RECENT_BINARY + 1: \
		case ID_RECENT_BINARY + 2: \
		case ID_RECENT_BINARY + 3: \
		case ID_RECENT_BINARY + 4: \
		case ID_RECENT_BINARY + 5: \
		case ID_RECENT_BINARY + 6: \
		case ID_RECENT_BINARY + 7: \
			no = LOWORD(wParam) - ID_RECENT_BINARY; \
			_tcscpy(path, config.recent_binary_path[drv][no]); \
			for(int i = no; i > 0; i--) { \
				_tcscpy(config.recent_binary_path[drv][i], config.recent_binary_path[drv][i - 1]); \
			} \
			_tcscpy(config.recent_binary_path[drv][0], path); \
			if(emu) { \
				emu->load_binary(drv, path); \
			} \
			break;
		BINARY_MENU_ITEMS(0, ID_LOAD_BINARY1, ID_SAVE_BINARY1, ID_RECENT_BINARY1)
#endif
#ifdef USE_BINARY_FILE2
		BINARY_MENU_ITEMS(1, ID_LOAD_BINARY2, ID_SAVE_BINARY2, ID_RECENT_BINARY2)
#endif
#ifdef _BML3MK5
		case ID_SCREEN_REC_SIZE1:
		case ID_SCREEN_REC_SIZE2:
			if (emu) {
				no = LOWORD(wParam) - ID_SCREEN_REC_SIZE1;
				emu->resize_rec_video(no);
			}
			break;
#endif
		case ID_SCREEN_REC60:
		case ID_SCREEN_REC30:
		case ID_SCREEN_REC20:
		case ID_SCREEN_REC15:
		case ID_SCREEN_REC12:
		case ID_SCREEN_REC10:
			if(emu) {
				emu->set_pause(1, true);
				rec_fps_no = LOWORD(wParam) - ID_SCREEN_REC60;
				emu->start_rec_video(fps[rec_fps_no], true);
				emu->start_rec_sound();
				fskip_remain = 0;
				emu->set_pause(1, false);
			}
			break;
		case ID_SCREEN_STOP:
			if(emu) {
				emu->stop_rec_video();
				emu->stop_rec_sound();
			}
			break;
		case ID_SCREEN_CAPTURE:
			if(emu) {
				emu->capture_screen();
			}
			break;
#ifdef _BML3MK5
		case ID_SCREEN_VFRAME:
			config.fps_no = -1;
			break;
		case ID_SCREEN_FPS60:
		case ID_SCREEN_FPS30:
		case ID_SCREEN_FPS20:
		case ID_SCREEN_FPS15:
		case ID_SCREEN_FPS12:
		case ID_SCREEN_FPS10:
			if(emu) {
				emu->set_pause(1, true);
				config.fps_no = LOWORD(wParam) - ID_SCREEN_FPS60;
				fskip_remain = 0;
				emu->set_pause(1, false);
			}
			break;
#endif
		case ID_SCREEN_WINDOW1:
		case ID_SCREEN_WINDOW2:
		case ID_SCREEN_WINDOW3:
		case ID_SCREEN_WINDOW4:
		case ID_SCREEN_WINDOW5:
		case ID_SCREEN_WINDOW6:
		case ID_SCREEN_WINDOW7:
		case ID_SCREEN_WINDOW8:
			if(emu) {
//				emu->set_window(hWnd, LOWORD(wParam) - ID_SCREEN_WINDOW1);
				emu->change_screen_mode(hWnd, LOWORD(wParam) - ID_SCREEN_WINDOW1);
			}
			break;
		case ID_SCREEN_FULLSCREEN1:
		case ID_SCREEN_FULLSCREEN2:
		case ID_SCREEN_FULLSCREEN3:
		case ID_SCREEN_FULLSCREEN4:
		case ID_SCREEN_FULLSCREEN5:
		case ID_SCREEN_FULLSCREEN6:
		case ID_SCREEN_FULLSCREEN7:
		case ID_SCREEN_FULLSCREEN8:
		case ID_SCREEN_FULLSCREEN9:
		case ID_SCREEN_FULLSCREEN10:
		case ID_SCREEN_FULLSCREEN11:
		case ID_SCREEN_FULLSCREEN12:
		case ID_SCREEN_FULLSCREEN13:
		case ID_SCREEN_FULLSCREEN14:
		case ID_SCREEN_FULLSCREEN15:
		case ID_SCREEN_FULLSCREEN16:
		case ID_SCREEN_FULLSCREEN17:
		case ID_SCREEN_FULLSCREEN18:
		case ID_SCREEN_FULLSCREEN19:
		case ID_SCREEN_FULLSCREEN20:
		case ID_SCREEN_FULLSCREEN21:
		case ID_SCREEN_FULLSCREEN22:
		case ID_SCREEN_FULLSCREEN23:
		case ID_SCREEN_FULLSCREEN24:
			if(emu && !emu->is_fullscreen()) {
//				emu->set_window(hWnd, LOWORD(wParam) - ID_SCREEN_FULLSCREEN1 + 8);
				emu->change_screen_mode(hWnd, LOWORD(wParam) - ID_SCREEN_FULLSCREEN1 + 8);
			}
			break;
		case ID_SCREEN_STRETCH:
//			config.stretch_screen = !config.stretch_screen;
			if(emu) {
//				emu->set_display_size(-1, -1, 10, !emu->is_fullscreen());
				emu->change_stretch_screen();
			}
			break;
		case ID_SCREEN_CUTOUT:
//			config.cutout_screen = !config.cutout_screen;
			if(emu) {
//				emu->set_display_size(-1, -1, 10, !emu->is_fullscreen());
				emu->change_cutout_screen();
			}
			break;
		// accelerator
		case ID_ACCEL_SCREEN:
			if(emu) {
//				emu->mute_sound(true);
//				emu->set_pause(1, true);
//				emu->set_window(hWnd, emu->is_fullscreen() ? prev_window_mode : -1);
				emu->change_screen_mode(hWnd, -1);
			}
			break;
		case ID_ACCEL_MOUSE:
			if(emu) {
				emu->toggle_mouse();
			}
			break;
#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
		case ID_SCREEN_MONITOR_TYPE0:
		case ID_SCREEN_MONITOR_TYPE1:
		case ID_SCREEN_MONITOR_TYPE2:
		case ID_SCREEN_MONITOR_TYPE3:
			config.monitor_type = LOWORD(wParam) - ID_SCREEN_MONITOR_TYPE0;
			if(emu) {
				emu->update_config();
#ifdef USE_SCREEN_ROTATE
//				if(emu->is_fullscreen()) {
//					emu->set_display_size(-1, -1, false);
//				}
//				else {
//					emu->set_window(hWnd, prev_window_mode);
//				}
				emu->change_screen_mode(hWnd, -1);
#endif
			}
			break;
#endif
#ifdef USE_SCANLINE
		case ID_SCREEN_SCANLINE:
			if(emu) {
				emu->change_screen_scanline();
			}
			break;
#endif
#ifdef _BML3MK5
		case ID_SCREEN_D3D_SYNC:
			if(emu) {
				emu->change_screen_use_direct3d(1);
			}
			break;
		case ID_SCREEN_D3D_ASYNC:
			if(emu) {
				emu->change_screen_use_direct3d(2);
			}
			break;
		case ID_SCREEN_D3D_A:
			if(emu) {
				emu->change_screen_use_direct3d(-1);
			}
			break;
		case ID_SCREEN_AFTERIMAGE1:
			if(emu) {
				emu->change_screen_afterimage(1);
			}
			break;
		case ID_SCREEN_AFTERIMAGE2:
			if(emu) {
				emu->change_screen_afterimage(2);
			}
			break;
		case ID_SCREEN_AFTERIMAGE_A:
			if(emu) {
				emu->change_screen_afterimage(-1);
			}
			break;
		case ID_SCREEN_KEEPIMAGE1:
			if(emu) {
				emu->change_screen_keepimage(1);
			}
			break;
		case ID_SCREEN_KEEPIMAGE2:
			if(emu) {
				emu->change_screen_keepimage(2);
			}
			break;
		case ID_SCREEN_KEEPIMAGE_A:
			if(emu) {
				emu->change_screen_keepimage(-1);
			}
			break;
#endif
		case ID_SOUND_REC:
			if(emu) {
				emu->start_rec_sound();
			}
			break;
		case ID_SOUND_STOP:
			if(emu) {
				emu->stop_rec_sound();
			}
			break;
		case ID_SOUND_VOLUME:
			if(emu) {
				emu->create_volume_dialog();
			}
			break;
		case ID_SOUND_FREQ0:
		case ID_SOUND_FREQ1:
		case ID_SOUND_FREQ2:
		case ID_SOUND_FREQ3:
		case ID_SOUND_FREQ4:
		case ID_SOUND_FREQ5:
		case ID_SOUND_FREQ6:
		case ID_SOUND_FREQ7:
			config.sound_frequency = LOWORD(wParam) - ID_SOUND_FREQ0;
			if(emu) {
				emu->update_config();
			}
			break;
		case ID_SOUND_LATE0:
		case ID_SOUND_LATE1:
		case ID_SOUND_LATE2:
		case ID_SOUND_LATE3:
		case ID_SOUND_LATE4:
			config.sound_latency = LOWORD(wParam) - ID_SOUND_LATE0;
			if(emu) {
				emu->update_config();
			}
			break;
#ifdef USE_SOUND_DEVICE_TYPE
		case ID_SOUND_DEVICE_TYPE0:
		case ID_SOUND_DEVICE_TYPE1:
		case ID_SOUND_DEVICE_TYPE2:
		case ID_SOUND_DEVICE_TYPE3:
			config.sound_device_type = LOWORD(wParam) - ID_SOUND_DEVICE_TYPE0;
			//if(emu) {
			//	emu->update_config();
			//}
			break;
#endif
#ifdef USE_BUTTON
		case ID_BUTTON +  0:
		case ID_BUTTON +  1:
		case ID_BUTTON +  2:
		case ID_BUTTON +  3:
		case ID_BUTTON +  4:
		case ID_BUTTON +  5:
		case ID_BUTTON +  6:
		case ID_BUTTON +  7:
		case ID_BUTTON +  8:
		case ID_BUTTON +  9:
		case ID_BUTTON + 10:
		case ID_BUTTON + 11:
		case ID_BUTTON + 12:
		case ID_BUTTON + 13:
		case ID_BUTTON + 14:
		case ID_BUTTON + 15:
		case ID_BUTTON + 16:
		case ID_BUTTON + 17:
		case ID_BUTTON + 18:
		case ID_BUTTON + 19:
		case ID_BUTTON + 20:
		case ID_BUTTON + 21:
		case ID_BUTTON + 22:
		case ID_BUTTON + 23:
		case ID_BUTTON + 24:
		case ID_BUTTON + 25:
		case ID_BUTTON + 26:
		case ID_BUTTON + 27:
		case ID_BUTTON + 28:
		case ID_BUTTON + 29:
		case ID_BUTTON + 30:
		case ID_BUTTON + 31:
			if(emu) {
				emu->press_button(LOWORD(wParam) - ID_BUTTON);
			}
			break;
#endif
#ifdef USE_PRINTER
		#define PRINTER_MENU_ITEMS(drv, ID_PRINTER_SAVE, ID_PRINTER_PRINT, ID_PRINTER_CLEAR, ID_PRINTER_DIRECT) \
		case ID_PRINTER_SAVE: \
			if(emu) { \
				open_printer_dialog(hWnd, drv); \
			} \
			break; \
		case ID_PRINTER_PRINT: \
			if(emu) { \
				emu->print_printer(drv); \
			} \
		break; \
		case ID_PRINTER_CLEAR: \
			if(emu) { \
				emu->clear_printer(drv); \
			} \
			break; \
		case ID_PRINTER_DIRECT: \
			if(emu) { \
				emu->enable_printer_direct(drv); \
			} \
			break;
		PRINTER_MENU_ITEMS(0, ID_PRINTER0_SAVE, ID_PRINTER0_PRINT, ID_PRINTER0_CLEAR, ID_PRINTER0_DIRECT)
		PRINTER_MENU_ITEMS(1, ID_PRINTER1_SAVE, ID_PRINTER1_PRINT, ID_PRINTER1_CLEAR, ID_PRINTER1_DIRECT)
		PRINTER_MENU_ITEMS(2, ID_PRINTER2_SAVE, ID_PRINTER2_PRINT, ID_PRINTER2_CLEAR, ID_PRINTER2_DIRECT)
#endif
#ifdef _BML3MK5
		#define COMM_MENU_ITEMS(drv, ID_COMM_SERVER, ID_COMM_CONNECT) \
		case ID_COMM_SERVER: \
			if(emu) { \
				emu->enable_comm_server(drv); \
			} \
			break; \
		case ID_COMM_CONNECT: \
			if(emu) { \
				emu->enable_comm_connect(drv); \
			} \
			break;
		COMM_MENU_ITEMS(0, ID_COMM0_SERVER, ID_COMM0_CONNECT)
		COMM_MENU_ITEMS(1, ID_COMM1_SERVER, ID_COMM1_CONNECT)

		case ID_OPTIONS_LEDBOX:
			if(emu) {
				emu->show_led();
			}
			break;
		case ID_OPTIONS_LEDBOX_INSIDE:
			if(emu) {
				emu->inside_led();
			}
			break;
		case ID_OPTIONS_MSGBOARD:
			if(emu) {
				emu->show_message_board();
			}
			break;
		case ID_OPTIONS_JOYPAD:
			if(emu) {
				emu->toggle_use_joypad();
			}
			break;
		case ID_OPTIONS_KEYBIND:
			if(emu) {
				emu->show_keybind_dialog();
			}
			break;
		case ID_OPTIONS_CONFIG:
			if(emu) {
				emu->show_config_dialog();
			}
			break;
#endif
		case ID_HELP_ABOUT:
			if(emu) emu->set_pause(1, true);
			DialogBox(hInst, MAKEINTRESOURCE(IDD_ABOUTBOX), hWnd, AboutWndProc);
			if(emu) emu->set_pause(1, false);
			break;
		}
		break;
	}
	return DefWindowProc(hWnd, iMsg, wParam, lParam) ;
}

static INT_PTR CALLBACK AboutWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	switch (message)
	{
	case WM_INITDIALOG:
		return (INT_PTR)TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
		{
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}

#ifdef USE_BUTTON
LRESULT CALLBACK ButtonWndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	for(int i = 0; i < MAX_BUTTONS; i++) {
		if(hWnd == hButton[i]) {
			switch(iMsg) {
			case WM_KEYDOWN:
			case WM_SYSKEYDOWN:
				if(emu) {
					emu->key_down(LOBYTE(wParam), false);
				}
				return 0;
			case WM_KEYUP:
			case WM_SYSKEYUP:
				if(emu) {
					emu->key_up(LOBYTE(wParam));
				}
				return 0;
			}
			return CallWindowProc(buttonWndProc[i], hWnd, iMsg, wParam, lParam);
		}
	}
	return 0;
}
#endif

void update_menu(HWND hWnd, HMENU hMenu, int pos)
{
	if (hMenu != GetSubMenu(GetMenu(hWnd), pos)) return;

#ifdef MENU_POS_CONTROL
	if(pos == MENU_POS_CONTROL) {
		// control menu
#ifdef USE_DIPSWITCH
		for(int i = 0; i < 8; i++) {
#ifdef _BML3MK5
			CheckMenuItem(hMenu, ID_DIPSWITCH1 + i, (config.dipswitch & (1 << i)) ? MF_CHECKED : MF_UNCHECKED);
#else
			CheckMenuItem(hMenu, ID_DIPSWITCH1 + i, !(config.dipswitch & (1 << i)) ? MF_CHECKED : MF_UNCHECKED);
#endif
		}
#endif
#ifdef _HC80
		if(config.device_type >= 0 && config.device_type < 3) {
			CheckMenuRadioItem(hMenu, ID_HC80_RAMDISK0, ID_HC80_RAMDISK2, ID_HC80_RAMDISK0 + config.device_type, MF_BYCOMMAND);
		}
#endif
#ifdef _MZ800
		if(config.boot_mode >= 0 && config.boot_mode < 2) {
			CheckMenuRadioItem(hMenu, ID_MZ800_MODE_MZ800, ID_MZ800_MODE_MZ700, ID_MZ800_MODE_MZ800 + config.boot_mode, MF_BYCOMMAND);
		}
#endif
#ifdef _PASOPIA
		if(config.boot_mode >= 0 && config.boot_mode < 5) {
			CheckMenuRadioItem(hMenu, ID_PASOPIA_MODE_TBASIC_V1_0, ID_PASOPIA_MODE_MINI_PASCAL, ID_PASOPIA_MODE_TBASIC_V1_0 + config.boot_mode, MF_BYCOMMAND);
		}
		if(config.device_type >= 0 && config.boot_mode < 3) {
			CheckMenuRadioItem(hMenu, ID_PASOPIA_DEVICE_RAM_PAC, ID_PASOPIA_DEVICE_JOYSTICK, ID_PASOPIA_DEVICE_RAM_PAC + config.device_type, MF_BYCOMMAND);
		}
#endif
#ifdef _PC98DO
		if(config.boot_mode >= 0 && config.boot_mode < 5) {
			CheckMenuRadioItem(hMenu, ID_PC98DO_MODE_PC98, ID_PC8801_MODE_N, ID_PC98DO_MODE_PC98 + config.boot_mode, MF_BYCOMMAND);
		}
#endif
#ifdef _PC8801MA
		if(config.boot_mode >= 0 && config.boot_mode < 4) {
			CheckMenuRadioItem(hMenu, ID_PC8801_MODE_V1S, ID_PC8801_MODE_N, ID_PC8801_MODE_V1S + config.boot_mode, MF_BYCOMMAND);
		}
		if(config.device_type >= 0 && config.boot_mode < 3) {
//			CheckMenuRadioItem(hMenu, ID_PC8801_DEVICE_JOYSTICK, ID_PC8801_DEVICE_JOYMOUSE, ID_PC8801_DEVICE_JOYSTICK + config.boot_mode, MF_BYCOMMAND);
			// joymouse is not supported yet...
			CheckMenuRadioItem(hMenu, ID_PC8801_DEVICE_JOYSTICK, ID_PC8801_DEVICE_MOUSE, ID_PC8801_DEVICE_JOYSTICK + config.device_type, MF_BYCOMMAND);
		}
#endif
#if defined(_PC9801E) || defined(_PC9801VM) || defined(_PC98DO) || defined(_PC8801MA)
		if(config.cpu_clock_low) {
			CheckMenuRadioItem(hMenu, ID_PC9801_CPU_CLOCK_HIGH, ID_PC9801_CPU_CLOCK_LOW, ID_PC9801_CPU_CLOCK_LOW, MF_BYCOMMAND);
		}
		else {
			CheckMenuRadioItem(hMenu, ID_PC9801_CPU_CLOCK_HIGH, ID_PC9801_CPU_CLOCK_LOW, ID_PC9801_CPU_CLOCK_HIGH, MF_BYCOMMAND);
		}
#endif
		if(config.cpu_power >= 0 && config.cpu_power <= 5) {
			CheckMenuRadioItem(hMenu, ID_CPU_POWER0, ID_CPU_POWER5, ID_CPU_POWER0 + config.cpu_power, MF_BYCOMMAND);
		}
#ifdef _BML3MK5
		CheckMenuItem(hMenu, ID_PAUSE,    emu->get_pause(2) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SYNC_IRQ, config.sync_irq ? MF_CHECKED : MF_UNCHECKED);
#endif
#ifdef USE_AUTO_KEY
		// auto key
		bool now_paste = true, now_stop = true;
		if(emu) {
			now_paste = emu->now_auto_key();
//			now_stop = !now_paste;
			now_stop = false;
		}
		EnableMenuItem(hMenu, ID_AUTOKEY_START, now_paste ? MF_GRAYED : MF_ENABLED);
		EnableMenuItem(hMenu, ID_AUTOKEY_STOP, now_stop ? MF_GRAYED : MF_ENABLED);
#endif
	}
#endif
#ifdef MENU_POS_CART
	else if(pos == MENU_POS_CART) {
		// cartridge
		bool flag = false;
		for(int i = 0; i < 8; i++) {
			DeleteMenu(hMenu, ID_RECENT_CART + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < 8; i++) {
			if(_tcscmp(config.recent_cart_path[i], _T(""))) {
				AppendMenu(hMenu, MF_STRING, ID_RECENT_CART + i, config.recent_cart_path[i]);
				flag = true;
			}
		}
		if(!flag) {
			AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_CART, _T("None"));
		}
	}
#endif
#define UPDATE_MENU_FD(drv, ID_OPEN_FD, ID_CHANGE_FD, ID_RECENT_FD, ID_D88_FILE_PATH, ID_SELECT_D88_BANK) { \
	bool flag = false; \
	while(DeleteMenu(hMenu, 4, MF_BYPOSITION) != 0) {} \
	if(d88_file[drv].bank_num > 1) { \
		AppendMenu(hMenu, MF_STRING | MF_DISABLED, ID_D88_FILE_PATH, d88_file[drv].path); \
		for(int i = 0; i < d88_file[drv].bank_num; i++) { \
			_TCHAR tmp[32]; \
			_stprintf(tmp, _T("%d: %s"), i + 1, d88_file[drv].bank[i].name); \
			AppendMenu(hMenu, MF_STRING | (i == d88_file[drv].cur_bank ? MF_CHECKED : 0), ID_SELECT_D88_BANK + i, tmp); \
		} \
		AppendMenu(hMenu, MF_SEPARATOR, 0, NULL); \
	} \
	for(int i = 0; i < 8; i++) { \
		if(_tcscmp(config.recent_disk_path[drv][i], _T(""))) { \
			AppendMenu(hMenu, MF_STRING, ID_RECENT_FD + i, config.recent_disk_path[drv][i]); \
			flag = true; \
		} \
	} \
	if(!flag) { \
		AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_FD, _T("None")); \
	} \
	set_disk_side_menu(hMenu, ID_CHANGE_FD, drv); \
	set_disk_file_menu(hMenu, ID_OPEN_FD, drv); \
}
#ifdef MENU_POS_FD1
	else if(pos == MENU_POS_FD1) {
		// floppy drive #1
		UPDATE_MENU_FD(0, ID_OPEN_FD1, ID_CHANGE_FD1, ID_RECENT_FD1, ID_D88_FILE_PATH1, ID_SELECT_D88_BANK1)
	}
#endif
#ifdef MENU_POS_FD2
	else if(pos == MENU_POS_FD2) {
		// floppy drive #2
		UPDATE_MENU_FD(1, ID_OPEN_FD2, ID_CHANGE_FD2, ID_RECENT_FD2, ID_D88_FILE_PATH2, ID_SELECT_D88_BANK2)
	}
#endif
#ifdef MENU_POS_FD3
	else if(pos == MENU_POS_FD3) {
		// floppy drive #3
		UPDATE_MENU_FD(2, ID_OPEN_FD3, ID_CHANGE_FD3, ID_RECENT_FD3, ID_D88_FILE_PATH3, ID_SELECT_D88_BANK3)
	}
#endif
#ifdef MENU_POS_FD4
	else if(pos == MENU_POS_FD4) {
		// floppy drive #4
		UPDATE_MENU_FD(3, ID_OPEN_FD4, ID_CHANGE_FD4, ID_RECENT_FD4, ID_D88_FILE_PATH4, ID_SELECT_D88_BANK4)
	}
#endif
#ifdef MENU_POS_FD5
	else if(pos == MENU_POS_FD5) {
		// floppy drive #5
		UPDATE_MENU_FD(4, ID_OPEN_FD5, ID_CHANGE_FD5, ID_RECENT_FD5, ID_D88_FILE_PATH5, ID_SELECT_D88_BANK5)
	}
#endif
#ifdef MENU_POS_FD6
	else if(pos == MENU_POS_FD6) {
		// floppy drive #6
		UPDATE_MENU_FD(5, ID_OPEN_FD6, ID_CHANGE_FD6, ID_RECENT_FD6, ID_D88_FILE_PATH6, ID_SELECT_D88_BANK6)
	}
#endif
#ifdef MENU_POS_QUICKDISK
	else if(pos == MENU_POS_QUICKDISK) {
		// quick disk drive
		bool flag = false;
		for(int i = 0; i < 8; i++) {
			DeleteMenu(hMenu, ID_RECENT_QUICKDISK + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < 8; i++) {
			if(_tcscmp(config.recent_quickdisk_path[i], _T(""))) {
				AppendMenu(hMenu, MF_STRING, ID_RECENT_QUICKDISK + i, config.recent_quickdisk_path[i]);
				flag = true;
			}
		}
		if(!flag) {
			AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_QUICKDISK, _T("None"));
		}
	}
#endif
#ifdef MENU_POS_DATAREC
	else if(pos == MENU_POS_DATAREC) {
		// data recorder
		bool flag = false;
		for(int i = 0; i < 8; i++) {
			DeleteMenu(hMenu, ID_RECENT_DATAREC + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < 8; i++) {
			if(_tcscmp(config.recent_datarec_path[i], _T(""))) {
				AppendMenu(hMenu, MF_STRING, ID_RECENT_DATAREC + i, config.recent_datarec_path[i]);
				flag = true;
			}
		}
		if(!flag) {
			AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_DATAREC, _T("None"));
		}

		set_datarec_file_menu(hMenu, ID_PLAY_DATAREC);
		// realmode
		CheckMenuItem(hMenu, ID_REAL_DATAREC, config.realmode_datarec ? MF_CHECKED : MF_UNCHECKED);
	}
#endif
#ifdef MENU_POS_MEDIA
	else if(pos == MENU_POS_MEDIA) {
		// media
		bool flag = false;
		for(int i = 0; i < 8; i++) {
			DeleteMenu(hMenu, ID_RECENT_MEDIA + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < 8; i++) {
			if(_tcscmp(config.recent_media_path[i], _T(""))) {
				AppendMenu(hMenu, MF_STRING, ID_RECENT_MEDIA + i, config.recent_media_path[i]);
				flag = true;
			}
		}
		if(!flag) {
			AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_MEDIA, _T("None"));
		}
	}
#endif
#ifdef MENU_POS_BINARY1
	else if(pos == MENU_POS_BINARY1) {
		// binary #1
		#define UPDATE_MENU_BINARY(drv, ID_RECENT_BINARY) \
		bool flag = false; \
		for(int i = 0; i < 8; i++) { \
			DeleteMenu(hMenu, ID_RECENT_BINARY + i, MF_BYCOMMAND); \
		} \
		for(int i = 0; i < 8; i++) { \
			if(_tcscmp(config.recent_binary_path[drv][i], _T(""))) { \
				AppendMenu(hMenu, MF_STRING, ID_RECENT_BINARY + i, config.recent_binary_path[drv][i]); \
				flag = true; \
			} \
		} \
		if(!flag) { \
			AppendMenu(hMenu, MF_GRAYED | MF_STRING, ID_RECENT_BINARY, _T("None")); \
		}
		UPDATE_MENU_BINARY(0, ID_RECENT_BINARY1)
		}
#endif
#ifdef MENU_POS_BINARY2
	else if(pos == MENU_POS_BINARY2) {
		// binary #2
		UPDATE_MENU_BINARY(1, ID_RECENT_BINARY2)
	}
#endif
#ifdef MENU_POS_SCREEN
	else if(pos == MENU_POS_SCREEN) {
		// recording
		bool now_rec = true, now_stop = true;
		if(emu) {
			now_rec = emu->now_rec_video();
			now_stop = !now_rec;
		}
#ifdef _BML3MK5
		// rec video size
		InsertMenu(hMenu, ID_SCREEN_REC_SIZE1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
		for(int i = 0; i <= (ID_SCREEN_REC_SIZE2 - ID_SCREEN_REC_SIZE1); i++) {
			DeleteMenu(hMenu, ID_SCREEN_REC_SIZE1 + i, MF_BYCOMMAND);
		}
		for(int i = 0; i <= (ID_SCREEN_REC_SIZE2 - ID_SCREEN_REC_SIZE1); i++) {
			_TCHAR buf[128];
			emu->get_screen_record_size_str(i, buf);
			InsertMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND | MF_STRING, ID_SCREEN_REC_SIZE1 + i, buf);
		}
		DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);
		CheckMenuRadioItem(hMenu, ID_SCREEN_REC_SIZE1, ID_SCREEN_REC_SIZE2, ID_SCREEN_REC_SIZE1 + config.screen_video_size, MF_BYCOMMAND);
		for(UINT i = ID_SCREEN_REC_SIZE1; i<= ID_SCREEN_REC_SIZE2; i++) {
			EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
		}
#endif
		for(UINT i = ID_SCREEN_REC60; i<= ID_SCREEN_REC10; i++) {
			EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
		}
		EnableMenuItem(hMenu, ID_SCREEN_STOP, now_stop ? MF_GRAYED : MF_ENABLED);

#ifdef _BML3MK5
		// frame rate
		CheckMenuRadioItem(hMenu, ID_SCREEN_VFRAME, ID_SCREEN_FPS10, ID_SCREEN_VFRAME + config.fps_no + 1, MF_BYCOMMAND);
		for(UINT i = ID_SCREEN_VFRAME; i<= ID_SCREEN_FPS10; i++) {
			EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
		}
#endif

		// screen mode
		UINT last = ID_SCREEN_WINDOW1;
		InsertMenu(hMenu, ID_SCREEN_WINDOW1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
		for(int i = 0; i < WINDOW_MODE_MAX; i++) {
			DeleteMenu(hMenu, ID_SCREEN_WINDOW1 + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < emu->get_window_mode_count(); i++) {
			_TCHAR buf[128];
			emu->get_window_mode_str(i, buf);
			InsertMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND | MF_STRING, ID_SCREEN_WINDOW1 + i, buf);
			last = ID_SCREEN_WINDOW1 + i;
		}
		DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);

		InsertMenu(hMenu, ID_SCREEN_FULLSCREEN1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
		for(int i = 0; i < SCREEN_MODE_MAX; i++) {
			DeleteMenu(hMenu, ID_SCREEN_FULLSCREEN1 + i, MF_BYCOMMAND);
		}
		for(int i = 0; i < emu->get_screen_mode_count(); i++) {
			MENUITEMINFO info;
			ZeroMemory(&info, sizeof(info));
			info.cbSize = sizeof(info);
			_TCHAR buf[128];
			emu->get_screen_mode_str(i, buf);
			info.fMask = MIIM_TYPE | MIIM_ID;
			info.fType = MFT_STRING;
			info.wID = ID_SCREEN_FULLSCREEN1 + i;
			info.dwTypeData = buf;
			InsertMenuItem(hMenu, ID_SCREEN_BOTTOM, FALSE, &info);
			EnableMenuItem(hMenu, ID_SCREEN_FULLSCREEN1 + i, emu->is_fullscreen() ? MF_GRAYED : MF_ENABLED);
			last = ID_SCREEN_FULLSCREEN1 + i;
		}
		DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);

		if(config.window_mode >= 0 && config.window_mode < emu->get_window_mode_count()) {
			CheckMenuRadioItem(hMenu, ID_SCREEN_WINDOW1, last, ID_SCREEN_WINDOW1 + config.window_mode, MF_BYCOMMAND);
		}
		else if(config.window_mode >= 8 && config.window_mode < emu->get_screen_mode_count() + 8) {
			CheckMenuRadioItem(hMenu, ID_SCREEN_FULLSCREEN1, last, ID_SCREEN_FULLSCREEN1 + config.window_mode - 8, MF_BYCOMMAND);
		}
		// stretch screen
		CheckMenuItem(hMenu, ID_SCREEN_STRETCH, config.stretch_screen ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_CUTOUT, config.cutout_screen ? MF_CHECKED : MF_UNCHECKED);

#ifdef USE_MONITOR_TYPE
		if(config.monitor_type >= 0 && config.monitor_type < USE_MONITOR_TYPE) {
			CheckMenuRadioItem(hMenu, ID_SCREEN_MONITOR_TYPE0, ID_SCREEN_MONITOR_TYPE0 + USE_MONITOR_TYPE - 1, ID_SCREEN_MONITOR_TYPE0 + config.monitor_type, MF_BYCOMMAND);
		}
#elif defined(USE_SCREEN_ROTATE)
		if(config.monitor_type >= 0 && config.monitor_type < 2) {
			CheckMenuRadioItem(hMenu, ID_SCREEN_MONITOR_TYPE0, ID_SCREEN_MONITOR_TYPE1, ID_SCREEN_MONITOR_TYPE0 + config.monitor_type, MF_BYCOMMAND);
		}
#endif
#ifdef USE_SCANLINE
		// scanline
		CheckMenuItem(hMenu, ID_SCREEN_SCANLINE, config.scan_line ? MF_CHECKED : MF_UNCHECKED);
#endif
#ifdef _BML3MK5
		CheckMenuItem(hMenu, ID_SCREEN_AFTERIMAGE1, (config.afterimage & 1) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_AFTERIMAGE2, (config.afterimage & 2) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_KEEPIMAGE1, (config.keepimage & 1) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_KEEPIMAGE2, (config.keepimage & 2) ? MF_CHECKED : MF_UNCHECKED);
#ifdef USE_DIRECT3D
		CheckMenuItem(hMenu, ID_SCREEN_D3D_SYNC, (config.use_direct3d & 1) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_D3D_ASYNC, (config.use_direct3d & 2) ? MF_CHECKED : MF_UNCHECKED);
		EnableMenuItem(hMenu, ID_SCREEN_D3D_SYNC, emu->enabled_direct3d() ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(hMenu, ID_SCREEN_D3D_ASYNC, emu->enabled_direct3d() ? MF_ENABLED : MF_GRAYED);
#else
		CheckMenuItem(hMenu, ID_SCREEN_D3D_SYNC, MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_SCREEN_D3D_ASYNC, MF_UNCHECKED);
		EnableMenuItem(hMenu, ID_SCREEN_D3D_SYNC, MF_GRAYED);
		EnableMenuItem(hMenu, ID_SCREEN_D3D_ASYNC, MF_GRAYED);
#endif
#ifdef USE_D3D
		CheckMenuItem(hMenu, ID_SCREEN_WAIT_VSYNC, config.wait_vsync ? MF_CHECKED : MF_UNCHECKED);
#endif
#endif
	}
#endif
#ifdef MENU_POS_SOUND
	else if(pos == MENU_POS_SOUND) {
		// sound menu
		bool now_rec = false, now_stop = false;
		if(emu) {
			now_rec = emu->now_rec_sound();
			now_stop = !now_rec;
		}
		EnableMenuItem(hMenu, ID_SOUND_REC, now_rec ? MF_GRAYED : MF_ENABLED);
		EnableMenuItem(hMenu, ID_SOUND_STOP, now_stop ? MF_GRAYED : MF_ENABLED);

		if(config.sound_frequency >= 0 && config.sound_frequency < 8) {
			CheckMenuRadioItem(hMenu, ID_SOUND_FREQ0, ID_SOUND_FREQ7, ID_SOUND_FREQ0 + config.sound_frequency, MF_BYCOMMAND);
		}
		if(config.sound_latency >= 0 && config.sound_latency < 5) {
			CheckMenuRadioItem(hMenu, ID_SOUND_LATE0, ID_SOUND_LATE4, ID_SOUND_LATE0 + config.sound_latency, MF_BYCOMMAND);
		}
#ifdef USE_SOUND_DEVICE_TYPE
		if(config.sound_device_type >= 0 && config.sound_device_type < USE_SOUND_DEVICE_TYPE) {
			CheckMenuRadioItem(hMenu, ID_SOUND_DEVICE_TYPE0, ID_SOUND_DEVICE_TYPE0 + USE_SOUND_DEVICE_TYPE - 1, ID_SOUND_DEVICE_TYPE0 + config.sound_device_type, MF_BYCOMMAND);
		}
#endif
	}
#endif
#ifdef MENU_POS_PRINTER
	else if(pos == MENU_POS_PRINTER) {
		for(int i=0; i<MAX_PRINTER; i++) {
			int j = i * 4;
			int printer_buffer_size = 0;
			// printer menu
			if(emu) {
				printer_buffer_size = emu->get_printer_buffer_size(i);
			}
			EnableMenuItem(hMenu, ID_PRINTER0_SAVE + j,  printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
			EnableMenuItem(hMenu, ID_PRINTER0_PRINT + j, printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
			EnableMenuItem(hMenu, ID_PRINTER0_CLEAR + j, printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
			CheckMenuItem(hMenu, ID_PRINTER0_DIRECT + j, config.printer_direct[i] ? MF_CHECKED : MF_UNCHECKED);
		}
	}
#endif
#ifdef MENU_POS_COMM
	else if(pos == MENU_POS_COMM) {
		for(int i=0; i<MAX_COMM; i++) {
			int j = i * 2;
			// comm menu
			CheckMenuItem(hMenu, ID_COMM0_SERVER + j, config.comm_server[i] ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_COMM0_CONNECT + j, config.comm_connect[i] ? MF_CHECKED : MF_UNCHECKED);
		}
	}
#endif
#ifdef MENU_POS_OPTIONS
	else if(pos == MENU_POS_OPTIONS) {
		// options menu
		if (emu) {
			CheckMenuItem(hMenu, ID_OPTIONS_LEDBOX, emu->is_shown_ledbox(0) ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_OPTIONS_LEDBOX_INSIDE, emu->is_shown_ledbox(1) ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_OPTIONS_MSGBOARD, emu->is_shown_message_board() ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_OPTIONS_JOYPAD, emu->is_enable_joypad() ? MF_CHECKED : MF_UNCHECKED);
		}
	}
#endif
	DrawMenuBar(hWnd);
}

#ifdef _BML3MK5
void execute_global_keys(HWND hWnd, int code, long status)
{
	switch(code) {
		case VK_RETURN:
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_ACCEL_SCREEN, 1), 0);
			break;
		case '0':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SYNC_IRQ, 1), 0);
			break;
		case '1':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER1, 1), 0);
			break;
		case '2':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER2, 1), 0);
			break;
		case '3':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER3, 1), 0);
			break;
		case '4':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER4, 1), 0);
			break;
		case '5':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER5, 1), 0);
			break;
		case '9':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER0, 1), 0);
			break;
		case 'C':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_CONFIG, 1), 0);
			break;
		case 'K':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_KEYBIND, 1), 0);
			break;
		case 'L':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_LEDBOX, 1), 0);
			break;
		case 'M':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_DIPSWITCH3, 1), 0);
			break;
		case 'P':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_RESET, 1), 0);
			break;
		case 'Q':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_PAUSE, 1), 0);
			break;
		case 'R':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_WARM_RESET, 1), 0);
			break;
		case 'S':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_SCANLINE, 1), 0);
			break;
		case 'T':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_AFTERIMAGE_A, 1), 0);
			break;
		case 'V':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SOUND_VOLUME, 1), 0);
			break;
		case 'Y':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_D3D_A, 1), 0);
			break;
		case 'Z':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_MSGBOARD, 1), 0);
			break;
		case VK_F4:
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_EXIT, 1), 0);
			break;
	}
}
#endif
