/** @file winutils.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2012.08.01 -

	@brief [ win32 utils ]
*/

#include "winutils.h"

extern EMU *emu;

// d88 bank switch
#ifdef USE_FD1
d88_file_t d88_file[8];

int inserted_disk_drive = 0;
#endif

#ifdef USE_DATAREC
int opened_tape_type = 0;
#endif

void get_long_full_path_name(_TCHAR* src, _TCHAR* dst)
{
	_TCHAR tmp[_MAX_PATH];

	if(GetFullPathName(src, _MAX_PATH, tmp, NULL) == 0) {
		_tcscpy(dst, src);
	}
	else if(GetLongPathName(tmp, dst, _MAX_PATH) == 0) {
		_tcscpy(dst, tmp);
	}
}

_TCHAR* get_parent_dir(_TCHAR* file)
{
	static _TCHAR path[_MAX_PATH];

	_tcscpy(path, file);
	int pt = (int)_tcslen(path);
	while(pt >= 0 && path[pt] != _T('\\')) {
		pt--;
	}
	path[pt + 1] = _T('\0');
	return path;
}

/// file dialog
_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, _TCHAR* ext, bool save, DWORD* flags)
{
	static _TCHAR path[_MAX_PATH];
	_TCHAR tmp[_MAX_PATH] = _T("");
	OPENFILENAME OpenFileName;
	int rc = 0;

	if (emu) emu->set_pause(1, true);

	memset(&OpenFileName, 0, sizeof(OpenFileName));
	OpenFileName.lStructSize = sizeof(OPENFILENAME);
	OpenFileName.hwndOwner = hWnd;
	OpenFileName.lpstrFilter = filter;
	OpenFileName.lpstrFile = tmp;
	OpenFileName.nMaxFile = _MAX_PATH;
	OpenFileName.lpstrTitle = title;
	if(ext != NULL && ext[0]) {
		OpenFileName.lpstrDefExt = ext;
	}
	if(dir != NULL && dir[0]) {
		OpenFileName.lpstrInitialDir = dir;
	}
	else {
		_TCHAR app[_MAX_PATH];
		GetModuleFileName(NULL, app, _MAX_PATH);
		OpenFileName.lpstrInitialDir = get_parent_dir(app);
	}
	if (save) {
		OpenFileName.Flags = OFN_OVERWRITEPROMPT;
		rc = GetSaveFileName(&OpenFileName);
	} else {
		OpenFileName.Flags = 0;
		rc = GetOpenFileName(&OpenFileName);
	}

	if (flags != NULL) *flags = OpenFileName.Flags;

	if (emu) emu->set_pause(1, false);

	if (rc) {
		get_long_full_path_name(OpenFileName.lpstrFile, path);
		_tcscpy(dir, get_parent_dir(path));
		return path;
	}
	return NULL;
}

_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, bool save)
{
	return get_open_file_name(hWnd, filter, title, dir, NULL, save, NULL);
}


bool check_file_extension(_TCHAR* filename, _TCHAR* ext)
{
	int nam_len = (int)_tcslen(filename);
	int ext_len = (int)_tcslen(ext);

	if(nam_len >= ext_len && _tcsncicmp(&filename[nam_len - ext_len], ext, ext_len) == 0) {
		return true;
	}
	return false;
}

////////////////////
#ifdef USE_CART
void open_cart_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
#ifdef _X1TWIN
		_T("Supported Files (*.pce)\0*.pce\0All Files (*.*)\0*.*\0\0"),
		_T("HuCARD"),
#else
		_T("Supported Files (*.rom;*.bin)\0*.rom;*.bin\0All Files (*.*)\0*.*\0\0"),
		_T("Game Cartridge"),
#endif
		config.initial_cart_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_cart_path);
		emu->open_cart(path);
	}
}
#endif

#ifdef USE_FD1
bool open_disk_dialog(HWND hWnd, int drv)
{
	DWORD flags;
	_TCHAR* path = get_open_file_name(
		hWnd,
#ifndef _BML3MK5
		_T("Supported Files (*.d88;*.d77;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)\0*.d88;*.d77;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7\0All Files (*.*)\0*.*\0\0"),
#else
		_T("Supported Files (*.d88;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)\0*.d88;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7\0All Files (*.*)\0*.*\0\0"),
#endif
		_T("Floppy Disk"),
		config.initial_disk_path,
		_T("d88"),
		false,
		&flags
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_disk_path[drv]);
		if (open_disk(drv, path, 0, flags)) {
			inserted_disk_drive |= (1 << drv);
			return true;
		} else {
			inserted_disk_drive &= ~(1 << drv);
			emu->out_log(LOG_WARN, _T("Floppy image couldn't be opend."));
			return false;
		}
	} else {
		return false;
	}
}

bool open_disk(int drv, _TCHAR* path, int bank, uint32 flags)
{
	bool rc = false;

	d88_file[drv].bank_num = 0;
	d88_file[drv].cur_bank = -1;
	d88_file[drv].bank[0].offset = 0;

	if(check_file_extension(path, _T(".d88")) || check_file_extension(path, _T(".d77"))) {
		FILE *fp = _tfopen(path, _T("rb"));
		if(fp != NULL) {
			try {
				fseek(fp, 0, SEEK_END);
				int file_size = ftell(fp), file_offset = 0;
				while(file_offset + 0x2b0 <= file_size && d88_file[drv].bank_num < MAX_D88_BANKS) {
					d88_file[drv].bank[d88_file[drv].bank_num].offset = file_offset;
					fseek(fp, file_offset, SEEK_SET);
#ifdef _UNICODE
					char tmp[18];
					fread(tmp, 17, 1, fp);
					tmp[17] = 0;
					MultiByteToWideChar(CP_ACP, 0, tmp, -1, d88_file[drv].bank[d88_file[drv].bank_num].name, 18);
#else
					fread(d88_file[drv].bank[d88_file[drv].bank_num].name, 17, 1, fp);
					d88_file[drv].bank[d88_file[drv].bank_num].name[17] = 0;
#endif
					fseek(fp, file_offset + 0x1c, SEEK_SET);
					file_offset += fgetc(fp);
					file_offset += fgetc(fp) << 8;
					file_offset += fgetc(fp) << 16;
					file_offset += fgetc(fp) << 24;
					d88_file[drv].bank_num++;
				}
				_tcscpy(d88_file[drv].path, path);
				d88_file[drv].cur_bank = bank;
			}
			catch(...) {
				d88_file[drv].bank_num = 0;
			}
		}
	}
	rc = emu->open_disk(drv, path, d88_file[drv].bank[bank].offset, flags);
	if (!rc) return rc;
#ifdef USE_FD2
	if(drv == 0 && bank + 1 < d88_file[drv].bank_num) {
		rc = open_disk(drv + 1, path, bank + 1, flags);
		if (!rc) return rc;
	}
#endif
#ifdef USE_FD4
	if(drv == 2 && bank + 1 < d88_file[drv].bank_num) {
		rc = open_disk(drv + 1, path, bank + 1, flags);
		if (!rc) return rc;
	}
#endif
#ifdef USE_FD6
	if(drv == 4 && bank + 1 < d88_file[drv].bank_num) {
		rc = open_disk(drv + 1, path, bank + 1, flags);
		if (!rc) return rc;
	}
#endif
	return rc;
}

void close_disk(int drv)
{
	emu->close_disk(drv);
	d88_file[drv].cur_bank = -1;
	inserted_disk_drive &= ~(1 << drv);
}

void set_disk_file_menu(HMENU hMenu, UINT uItem, int drv)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	if (inserted_disk_drive & (1 << drv)) {
		_tcscpy(str,_T("Open ["));
		_tcscat(str,config.recent_disk_path[drv][0]);
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Open..."));
	}

	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);

	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
}

#ifdef _BML3MK5
void set_disk_side_menu(HMENU hMenu, UINT uItem, int drv)
{
	if (emu) {
		if (config.fdd_type != 1) {
			EnableMenuItem(hMenu, uItem, MF_GRAYED);
		} else {
			MENUITEMINFO minfo;
			_TCHAR str[40];

			minfo.cbSize = sizeof(MENUITEMINFO);
			minfo.fMask = MIIM_TYPE;
			minfo.fType = MFT_STRING;

			int side = emu->get_disk_side(drv);

			if (side) {
				_tcscpy(str,_T("Change side to A"));
			} else {
				_tcscpy(str,_T("Change side to B"));
			}
			minfo.dwTypeData = str;
			minfo.cch = (UINT)_tcslen(str);

			SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
			EnableMenuItem(hMenu, uItem, MF_ENABLED);
		}
	}
}
#endif
#endif	// USE_FD1

#ifdef USE_QUICKDISK
void open_quickdisk_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.mzt)\0*.mzt\0All Files (*.*)\0*.*\0\0"),
		_T("Quick Disk"),
		config.initial_quickdisk_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_quickdisk_path);
		emu->open_quickdisk(path);
	}
}
#endif

#ifdef USE_DATAREC
void open_datarec_dialog(HWND hWnd, bool play)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
#if defined(DATAREC_PC8801)
		play ? _T("Supported Files (*.cas;*.cmt;*.t88)\0*.cas;*.cmt;*.t88\0All Files (*.*)\0*.*\0\0")
		     : _T("Supported Files (*.cas;*.cmt)\0*.cas;*.cmt\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_BINARY_ONLY)
		_T("Supported Files (*.cas;*.cmt)\0*.cas;*.cmt\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_TAP)
		play ? _T("Supported Files (*.wav;*.cas;*.tap)\0*.wav;*.cas;*.tap\0All Files (*.*)\0*.*\0\0")
		     : _T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_MZT)
		play ? _T("Supported Files (*.wav;*.cas;*.mzt;*.m12)\0*.wav;*.cas;*.mzt;*.m12\0All Files (*.*)\0*.*\0\0")
		     : _T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#else
#ifdef _BML3MK5
		(play ?
		_T("Supported Files (*.l3;*.l3b;*.l3c;*.wav)\0*.l3;*.l3b;*.l3c;*.wav\0All Files (*.*)\0*.*\0\0")
		:
		_T("L3 File (*.l3)\0*.l3\0L3B File (*.l3b)\0*.l3b\0L3C File (*.l3c)\0*.l3c\0Wave File (*.wav)\0*.wav\0All Files (*.*)\0*.*\0\0")
		),
#else
		_T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#endif
#endif
		_T("Data Recorder Tape"),
		config.initial_datarec_path,
		_T("l3"),
		(play ? false : true),
		NULL
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_datarec_path);
		if(play) {
			if (emu->play_datarec(path)) {
				opened_tape_type = 1;
			} else {
				emu->out_log(LOG_WARN, _T("Tape image couldn't be opend."));
			}
		}
		else {
			if (emu->rec_datarec(path)) {
				opened_tape_type = 2;
			} else {
				emu->out_log(LOG_ERROR, _T("Tape image couldn't be saved."));
			}
		}
	}
}

void set_datarec_file_menu(HMENU hMenu, UINT uItem)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	if (opened_tape_type == 1) {
		_tcscpy(str,_T("Play ["));
		_tcscat(str,config.recent_datarec_path[0]);
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Play..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);

	if (opened_tape_type == 2) {
		_tcscpy(str,_T("Rec ["));
		_tcscat(str,config.recent_datarec_path[0]);
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Rec..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem + 1, FALSE, &minfo);

	CheckMenuRadioItem(hMenu, uItem, uItem + 1, uItem + opened_tape_type - 1, MF_BYCOMMAND);
}
#endif


#ifdef USE_MEDIA
void open_media_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.m3u)\0*.m3u\0All Files (*.*)\0*.*\0\0"),
		_T("Sound Cassette Tape"),
		config.initial_media_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_media_path);
		emu->open_media(path);
	}
}
#endif

#ifdef USE_BINARY_FILE1
void open_binary_dialog(HWND hWnd, int drv, bool load)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.ram;*.bin)\0*.ram;*.bin\0All Files (*.*)\0*.*\0\0"),
#if defined(_PASOPIA) || defined(_PASOPIA7)
		_T("RAM Pack Cartridge"),
#else
		_T("Memory Dump"),
#endif
		config.initial_binary_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_binary_path[drv]);
		if(load) {
			emu->load_binary(drv, path);
		}
		else {
			emu->save_binary(drv, path);
		}
	}
}
#endif

#ifdef USE_PRINTER
void open_printer_dialog(HWND hWnd, int drv)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.lpt)\0*.lpt\0All Files (*.*)\0*.*\0\0"),
		_T("Print Data"),
		config.initial_printer_path,
		_T("lpt"),
		TRUE,
		NULL
		);
	if(path) {
		emu->save_printer(drv, path);
	}
}
#endif

#ifdef USE_AUTO_KEY
void open_autokey_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.txt;*.bas;*.lpt)\0*.txt;*.bas;*.lpt\0All Files (*.*)\0*.*\0\0"),
		_T("Open Auto key file"),
		config.initial_autokey_path,
		_T("txt"),
		false,
		NULL
		);
	if(path) {
		emu->start_auto_key(path);
	}
}
#endif

#ifdef _BML3MK5
void open_save_state_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3r)\0*.l3r\0All Files (*.*)\0*.*\0\0"),
		_T("Save Status Data"),
		config.initial_state_path,
		_T("l3r"),
		true,
		NULL
		);
	if(path) {
		emu->save_state(path);
	}
}

void open_load_state_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3r)\0*.l3r\0All Files (*.*)\0*.*\0\0"),
		_T("Load Status Data"),
		config.initial_state_path,
		_T("l3r"),
		false,
		NULL
		);
	if(path) {
		emu->load_state(path);
	}
}
#endif
