/** @file winutils.h

	Skelton for retropc emulator

	@author Sasaji
	@date   2012.08.01 -

	@brief [ win32 utils ]
*/

#ifndef _WINUTILS_H_
#define _WINUTILS_H_

#include <windows.h>
#include <stdio.h>
#include "emu.h"

#define UPDATE_HISTORY(path, recent) { \
	int no = 7; \
	for(int i = 0; i < 8; i++) { \
		if(_tcscmp(recent[i], path) == 0) { \
			no = i; \
			break; \
		} \
	} \
	for(int i = no; i > 0; i--) { \
		_tcscpy(recent[i], recent[i - 1]); \
	} \
	_tcscpy(recent[0], path); \
}

// d88 bank switch
#ifdef USE_FD1
#define MAX_D88_BANKS 50

typedef struct {
	_TCHAR name[18];
	int offset;
} d88_bank_t;
typedef struct {
	_TCHAR path[MAX_PATH];
	d88_bank_t bank[MAX_D88_BANKS];
	int bank_num;
	int cur_bank;
} d88_file_t;
extern d88_file_t d88_file[8];

extern int inserted_disk_drive;
#endif

#ifdef USE_DATAREC
extern int opened_tape_type;
#endif

// prototype
void    get_long_full_path_name(_TCHAR* src, _TCHAR* dst);
_TCHAR* get_parent_dir(_TCHAR* file);
_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, _TCHAR* ext = NULL, bool save = false, DWORD* flags = NULL);
_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, bool save);

bool    check_file_extension(_TCHAR* filename, _TCHAR* ext);

// dialog
#ifdef USE_CART
void open_cart_dialog(HWND hWnd);
#endif
#ifdef USE_FD1
bool open_disk_dialog(HWND hWnd, int drv);
bool open_disk(int drv, _TCHAR* filename, int bank, uint32 flags);
void close_disk(int drv);
void set_disk_file_menu(HMENU hMenu, UINT uItem, int drv);
#ifdef _BML3MK5
void set_disk_side_menu(HMENU hMenu, UINT uItem, int drv);
#endif
#endif
#ifdef USE_QUICKDISK
void open_quickdisk_dialog(HWND hWnd);
#endif
#ifdef USE_DATAREC
void open_datarec_dialog(HWND hWnd, bool play);
void set_datarec_file_menu(HMENU hMenu, UINT uItem);
#endif
#ifdef USE_MEDIA
void open_media_dialog(HWND hWnd);
#endif
#ifdef USE_BINARY_FILE1
void open_binary_dialog(HWND hWnd, int drv, bool load);
#endif
#ifdef USE_PRINTER
void open_printer_dialog(HWND hWnd, int drv);
#endif
#ifdef USE_AUTO_KEY
void open_autokey_dialog(HWND hWnd);
#endif
#ifdef _BML3MK5
void open_save_state_dialog(HWND hWnd);
void open_load_state_dialog(HWND hWnd);
#endif

#endif
